@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,47 +13,53 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
 ******************************************************************************/
 
-
 /******************************************************************************
 *               (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                  *
 *                                                                             *
 *                        coded by Peter Unold 1992                            *
+*                    recoded by Igor van den Hoven 2004                       *
 ******************************************************************************/
 
 #include "tintin.h"
 #include <sys/stat.h>
 
-/*
-	read and execute a command file, supports multi lines - Igor
-*/
 
 DO_COMMAND(do_read)
 {
 	FILE *fp;
-	struct stat filedata;
-	char *bufi, *bufo, filename[BUFFER_SIZE], temp[BUFFER_SIZE], *pti, *pto, last = 0;
-	int lvl, cnt, com, lnc, fix, ok;
-	int counter[LIST_MAX];
 
-	get_arg_in_braces(ses, arg, filename, TRUE);
-	substitute(ses, filename, filename, SUB_VAR|SUB_FUN);
+	sub_arg_in_braces(ses, arg, arg1, GET_ONE, SUB_VAR|SUB_FUN);
 
-	if ((fp = fopen(filename, "r")) == NULL)
+	if ((fp = fopen(arg1, "r")) == NULL)
 	{
-		tintin_printf(ses, "#READ {%s} - FILE NOT FOUND.", filename);
+		check_all_events(ses, SUB_ARG, 0, 2, "READ ERROR", arg1, "FILE NOT FOUND.");
+
+		tintin_printf(ses, "#READ {%s} - FILE NOT FOUND.", arg1);
+
 		return ses;
 	}
 
+	return read_file(ses, fp, arg1);
+}
+
+struct session *read_file(struct session *ses, FILE *fp, char *filename)
+{
+	char *bufi, *bufo, temp[INPUT_SIZE], *pti, *pto, last = 0;
+	int lvl, cnt, com, lnc, fix, ok, verbose, size;
+	int counter[LIST_MAX];
+
 	temp[0] = getc(fp);
 
-	if (!isgraph((int) temp[0]) || isalpha((int) temp[0]))
+	if (!ispunct((int) temp[0]))
 	{
-		tintin_printf(ses, "#ERROR: #READ {%s} - INVALID START OF FILE.", filename);
+		check_all_events(ses, SUB_ARG, 0, 2, "READ ERROR", filename, "INVALID START OF FILE");
+
+		tintin_printf(ses, "#ERROR: #READ {%s} - INVALID START OF FILE '%c'.", filename, temp[0]);
 
 		fclose(fp);
 
@@ -71,18 +76,25 @@ DO_COMMAND(do_read)
 		}
 	}
 
-	stat(filename, &filedata);
+	fseek(fp, 0, SEEK_END);
 
-	if ((bufi = (char *) calloc(1, filedata.st_size + 2)) == NULL || (bufo = (char *) calloc(1, filedata.st_size + 2)) == NULL)
+	size = ftell(fp);
+
+	fseek(fp, 0, SEEK_SET);
+
+	if ((bufi = (char *) calloc(1, size + 2)) == NULL || (bufo = (char *) calloc(1, size + 2)) == NULL)
 	{
-		tintin_printf(ses, "#ERROR: #READ {%s} - FAILED TO ALLOCATE MEMORY.", filename);
+		check_all_events(ses, SUB_ARG, 0, 2, "READ ERROR", filename, "FAILED TO ALLOCATE MEMORY");
+
+		tintin_printf(ses, "#ERROR: #READ {%s} - FAILED TO ALLOCATE %d BYTES OF MEMORY.", filename, size + 2);
 
 		fclose(fp);
 
 		return ses;
 	}
 
-	fread(bufi, 1, filedata.st_size, fp);
+
+	fread(bufi, 1, size, fp);
 
 	pti = bufi;
 	pto = bufo;
@@ -93,10 +105,11 @@ DO_COMMAND(do_read)
 	{
 		if (com == 0)
 		{
-			if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+			if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 			{
 				*pto++ = *pti++;
 				*pto++ = *pti++;
+
 				continue;
 			}
 
@@ -186,6 +199,14 @@ DO_COMMAND(do_read)
 
 						}
 
+						if (last == 0)
+						{
+							if (*pti != DEFAULT_OPEN && *pti != DEFAULT_CLOSE)
+							{
+								show_error(ses, LIST_COMMAND, "#WARNING: #READ {%s} MISSING SEMICOLON ON LINE %d.", filename, lnc - 1);
+							}
+						}
+
 						if (*pti != DEFAULT_CLOSE && last == 0)
 						{
 							*pto++ = ' ';
@@ -266,8 +287,11 @@ DO_COMMAND(do_read)
 	*pto++ = '\n';
 	*pto   = '\0';
 
+
 	if (lvl)
 	{
+		check_all_events(ses, SUB_ARG, 0, 2, "READ ERROR", filename, "MISSING BRACE OPEN OR CLOSE");
+
 		tintin_printf(ses, "#ERROR: #READ {%s} - MISSING %d '%c' BETWEEN LINE %d AND %d.", filename, abs(lvl), lvl < 0 ? DEFAULT_OPEN : DEFAULT_CLOSE, fix == 0 ? 1 : ok, fix == 0 ? lnc + 1 : fix);
 
 		fclose(fp);
@@ -280,6 +304,8 @@ DO_COMMAND(do_read)
 
 	if (com)
 	{
+		check_all_events(ses, SUB_ARG, 0, 2, "READ ERROR", filename, "MISSING COMMENT OPEN OR CLOSE");
+
 		tintin_printf(ses, "#ERROR: #READ {%s} - MISSING %d '%s'", filename, abs(com), com < 0 ? "/*" : "*/");
 
 		fclose(fp);
@@ -290,13 +316,33 @@ DO_COMMAND(do_read)
 		return ses;
 	}
 
-	sprintf(temp, "{TINTIN CHAR} {%c}", bufo[0]);
+	sprintf(temp, "#CONFIG {TINTIN CHAR} {%c}", bufo[0]);
 
-	gtd->quiet++;
+	if (bufo[0] != gtd->tintin_char)
+	{
+		gtd->level->verbose++;
+		gtd->level->debug++;
 
-	do_configure(ses, temp);
+		show_error(ses, LIST_COMMAND, "\e[1;5;31mWARNING: SETTING THE COMMAND CHARACTER TO '%c' BECAUSE IT'S THE FIRST CHARACTER IN THE FILE.", bufo[0]);
+
+		gtd->level->debug--;
+		gtd->level->verbose--;
+
+		gtd->level->quiet++;
+
+		script_driver(ses, LIST_COMMAND, temp);
+
+		gtd->level->quiet--;
+	}
+
+	verbose = HAS_BIT(ses->flags, SES_FLAG_VERBOSE) ? 1 : 0;
+
+	gtd->level->input++;
+
+	gtd->level->verbose += verbose;
 
 	lvl = 0;
+	lnc = 0;
 	pti = bufo;
 	pto = bufi;
 
@@ -307,24 +353,13 @@ DO_COMMAND(do_read)
 			*pto++ = *pti++;
 			continue;
 		}
+		lnc++;
 		*pto = 0;
 
-		if (strlen(bufi) >= BUFFER_SIZE)
+		if (pto - bufi >= BUFFER_SIZE)
 		{
-/*			gtd->quiet--;
-
-			bufi[20] = 0;
-*/
-//			tintin_printf(ses, "#ERROR: #READ {%s} - BUFFER OVERFLOW AT COMMAND: %.20s.", filename, bufi);
-			tintin_printf(ses, "#ERROR: #READ {%s} - BUFFER OVERFLOW AT COMMAND:", filename);
-/*
-			fclose(fp);
-
-			free(bufi);
-			free(bufo);
-
-			return ses;
-*/		}
+			show_debug(ses, LIST_COMMAND, "#WARNING: #READ {%s} - POSSIBLE BUFFER OVERFLOW AT COMMAND: %.30s", filename, bufi);
+		}
 
 		if (bufi[0])
 		{
@@ -334,7 +369,9 @@ DO_COMMAND(do_read)
 		pti++;
 	}
 
-	gtd->quiet--;
+	gtd->level->verbose -= verbose;
+
+	gtd->level->input--;
 
 	if (!HAS_BIT(ses->flags, SES_FLAG_VERBOSE))
 	{
@@ -370,15 +407,37 @@ DO_COMMAND(do_read)
 DO_COMMAND(do_write)
 {
 	FILE *file;
-	char filename[BUFFER_SIZE];
 	struct listroot *root;
-	int i, j, cnt = 0;
+	struct listnode *node;
 
-	get_arg_in_braces(ses, arg, filename, TRUE);
+	int i, j, fix, cnt = 0;
 
-	if (*filename == 0 || (file = fopen(filename, "w")) == NULL)
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = get_arg_in_braces(ses, arg, arg2, GET_ONE);
+
+	if (*arg1 == 0)
 	{
-		tintin_printf(ses, "#ERROR: #WRITE: COULDN'T OPEN {%s} TO WRITE.", filename);
+		check_all_events(ses, SUB_ARG, 0, 2, "WRITE ERROR", arg1, "INVALID FILE NAME");
+
+		tintin_printf2(ses, "#SYNTAX: #WRITE {<filename>} {[FORCE]}");
+
+		return ses;
+	}
+	
+	if (!str_suffix(arg1, ".map") && !is_abbrev(arg2, "FORCE"))
+	{
+		check_all_events(ses, SUB_ARG, 0, 2, "WRITE ERROR", arg1, "INVALID FILE EXTENSION");
+		tintin_printf2(ses, "#WRITE {%s}: USE {FORCE} TO OVERWRITE .map FILES.", arg1);
+
+		return ses;
+	}
+
+	if ((file = fopen(arg1, "w")) == NULL)
+	{
+		check_all_events(ses, SUB_ARG, 0, 2, "WRITE ERROR", arg1, "FAILED TO OPEN");
+
+		tintin_printf(ses, "#ERROR: #WRITE: COULDN'T OPEN {%s} TO WRITE.", arg1);
+
 		return ses;
 	}
 
@@ -391,20 +450,35 @@ DO_COMMAND(do_write)
 			continue;
 		}
 
+		fix = 0;
+
 		for (j = 0 ; j < root->used ; j++)
 		{
-			if (*root->list[j]->group == 0)
+			node = root->list[j];
+
+			if (node->shots)//HAS_BIT(node->flags, NODE_FLAG_ONESHOT))
 			{
-				write_node(ses, i, root->list[j], file);
+				continue;
+			}
+
+			if (*node->group == 0)
+			{
+				write_node(ses, i, node, file);
 
 				cnt++;
+				fix++;
 			}
 		}
+
+		if (fix)
+		{
+			fputs("\n", file);
+		}
 	}
 
 	fclose(file);
 
-	show_message(ses, LIST_COMMAND, "#WRITE: %d COMMANDS WRITTEN TO {%s}.", cnt, filename);
+	show_message(ses, LIST_COMMAND, "#WRITE: %d COMMANDS WRITTEN TO {%s}.", cnt, arg1);
 
 	return ses;
 }
@@ -414,8 +488,8 @@ void write_node(struct session *ses, int list, struct listnode *node, FILE *file
 {
 	char *result, *str;
 
-	int llen = UMAX(20, strlen(node->left));
-	int rlen = UMAX(25, strlen(node->right));
+	int llen = UMAX(20, strlen(node->arg1));
+	int rlen = UMAX(25, strlen(node->arg2));
 
 	push_call("write_node(%d,%p,%p)",list,node,file);
 
@@ -424,12 +498,20 @@ void write_node(struct session *ses, int list, struct listnode *node, FILE *file
 		case LIST_EVENT:
 		case LIST_FUNCTION:
 		case LIST_MACRO:
-			asprintf(&result, "%c%s {%s}\n{\n%s\n}\n\n", gtd->tintin_char, list_table[list].name, node->left, script_writer(ses, node->right));
+			asprintf(&result, "%c%s {%s}\n{\n%s\n}\n\n", gtd->tintin_char, list_table[list].name, node->arg1, script_writer(ses, node->arg2));
 			break;
 
 		case LIST_ACTION:
 		case LIST_ALIAS:
-			asprintf(&result, "%c%s {%s}\n{\n%s\n}\n{%s}\n\n", gtd->tintin_char, list_table[list].name, node->left, script_writer(ses, node->right), node->pr);
+		case LIST_BUTTON:
+			if (!strcmp(node->arg3, "5"))
+			{
+				asprintf(&result, "%c%s {%s}\n{\n%s\n}\n\n", gtd->tintin_char, list_table[list].name, node->arg1, script_writer(ses, node->arg2));
+			}
+			else
+			{
+				asprintf(&result, "%c%s {%s}\n{\n%s\n}\n{%s}\n\n", gtd->tintin_char, list_table[list].name, node->arg1, script_writer(ses, node->arg2), node->arg3);
+			}
 			break;
 
 		case LIST_VARIABLE:
@@ -437,7 +519,7 @@ void write_node(struct session *ses, int list, struct listnode *node, FILE *file
 
 			show_nest_node(node, &str, 1);
 
-			asprintf(&result, "%c%-16s {%s} %*s {%s}\n", gtd->tintin_char, list_table[list].name, node->left, 20 - llen, "", str);
+			asprintf(&result, "%c%-16s {%s} %*s {%s}\n", gtd->tintin_char, list_table[list].name, node->arg1, 20 - llen, "", str);
 
 			str_free(str);
 
@@ -450,13 +532,16 @@ void write_node(struct session *ses, int list, struct listnode *node, FILE *file
 					result = strdup("");
 					break;
 				case 1:
-					asprintf(&result, "%c%-16s {%s}\n", gtd->tintin_char, list_table[list].name, node->left);
+					asprintf(&result, "%c%-16s {%s}\n", gtd->tintin_char, list_table[list].name, node->arg1);
 					break;
 				case 2:
-					asprintf(&result, "%c%-16s {%s} %*s {%s}\n", gtd->tintin_char, list_table[list].name, node->left, 20 - llen, "", node->right);
+					asprintf(&result, "%c%-16s {%s} %*s {%s}\n", gtd->tintin_char, list_table[list].name, node->arg1, 20 - llen, "", node->arg2);
 					break;
 				case 3:
-					asprintf(&result, "%c%-16s {%s} %*s {%s} %*s {%s}\n", gtd->tintin_char, list_table[list].name, node->left, 20 - llen, "", node->right, 25 - rlen, "", node->pr);
+					asprintf(&result, "%c%-16s {%s} %*s {%s} %*s {%s}\n", gtd->tintin_char, list_table[list].name, node->arg1, 20 - llen, "", node->arg2, 25 - rlen, "", node->arg3);
+					break;
+				case 4:
+					asprintf(&result, "%c%-16s {%s} %*s {%s} %*s {%s} {%s}\n", gtd->tintin_char, list_table[list].name, node->arg1, 20 - llen, "", node->arg2, 25 - rlen, "", node->arg3, node->arg4);
 					break;
 			}
 			break;
