@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2006 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2020 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -15,12 +14,11 @@
 *   GNU General Public License for more details.                              *
 *                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
-*******************************************************************************/
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
+******************************************************************************/
 
 /******************************************************************************
-*                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
+*                               T I N T I N + +                               *
 *                                                                             *
 *                      coded by Igor van den Hoven 2006                       *
 ******************************************************************************/
@@ -29,16 +27,17 @@
 
 DO_COMMAND(do_history)
 {
-	char left[BUFFER_SIZE], right[BUFFER_SIZE];
 	int cnt;
 
-	arg = get_arg_in_braces(ses, arg, left,  FALSE);
+	arg = get_arg_in_braces(ses, arg, arg1, GET_ONE);
+	arg = sub_arg_in_braces(ses, arg, arg2, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg3, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg4, GET_ALL, SUB_VAR|SUB_FUN);
 
-	arg = get_arg_in_braces(ses, arg, right, TRUE);
-	substitute(ses, right, right, SUB_VAR|SUB_FUN);
-
-	if (*left == 0)
+	if (*arg1 == 0)
 	{
+		info:
+
 		tintin_header(ses, " HISTORY COMMANDS ");
 
 		for (cnt = 0 ; *history_table[cnt].name != 0 ; cnt++)
@@ -52,17 +51,17 @@ DO_COMMAND(do_history)
 
 	for (cnt = 0 ; *history_table[cnt].name ; cnt++)
 	{
-		if (!is_abbrev(left, history_table[cnt].name))
+		if (!is_abbrev(arg1, history_table[cnt].name))
 		{
 			continue;
 		}
 
-		history_table[cnt].fun(ses, right);
+		history_table[cnt].fun(ses, arg2, arg3, arg4);
 
 		return ses;
 	}
 
-	do_history(ses, "");
+	goto info;
 
 	return ses;
 }
@@ -74,7 +73,7 @@ void add_line_history(struct session *ses, char *line)
 
 	root = ses->list[LIST_HISTORY];
 
-	if (HAS_BIT(root->flags, LIST_FLAG_IGNORE))
+	if (HAS_BIT(root->flags, LIST_FLAG_IGNORE) || gtd->level->ignore)
 	{
 		return;
 	}
@@ -83,7 +82,7 @@ void add_line_history(struct session *ses, char *line)
 	{
 		if (root->used && HAS_BIT(ses->flags, SES_FLAG_REPEATENTER))
 		{
-			strcpy(line, root->list[root->used - 1]->left);
+			strcpy(line, root->list[root->used - 1]->arg1);
 		}
 		return;
 	}
@@ -93,7 +92,7 @@ void add_line_history(struct session *ses, char *line)
 		search_line_history(ses, line);
 	}
 
-	update_node_list(ses->list[LIST_HISTORY], line, "", "");
+	update_node_list(ses->list[LIST_HISTORY], line, "", "", "");
 
 	while (root->used > gtd->history_size)
 	{
@@ -112,9 +111,9 @@ void search_line_history(struct session *ses, char *line)
 
 	for (i = root->used - 1 ; i >= 0 ; i--)
 	{
-		if (!strncmp(root->list[i]->left, &line[1], strlen(&line[1])))
+		if (!strncmp(root->list[i]->arg1, &line[1], strlen(&line[1])))
 		{
-			strcpy(line, root->list[i]->left);
+			strcpy(line, root->list[i]->arg1);
 
 			return;
 		}
@@ -124,7 +123,7 @@ void search_line_history(struct session *ses, char *line)
 
 DO_HISTORY(history_character)
 {
-	gtd->repeat_char = *arg;
+	gtd->repeat_char = *arg1;
 
 	show_message(ses, LIST_HISTORY, "#HISTORY CHARACTER SET TO {%c}.", gtd->repeat_char);
 }
@@ -141,19 +140,77 @@ DO_HISTORY(history_delete)
 
 DO_HISTORY(history_insert)
 {
-	add_line_history(ses, arg);
+	add_line_history(ses, arg1);
+}
+
+DO_HISTORY(history_get)
+{
+	struct listroot *root = ses->list[LIST_HISTORY];
+	int cnt, min, max;
+
+	if (*arg1 == 0)
+	{
+		show_error(ses, LIST_COMMAND, "#SYNTAX: #HISTORY GET <VARIABLE> [LOWER BOUND] [UPPER BOUND]");
+
+		return;
+	}
+
+	min = get_number(ses, arg2);
+
+	if (min < 0)
+	{
+		min = root->used + min;
+	}
+
+	min = URANGE(0, min, root->used - 1);
+
+	if (*arg3 == 0)
+	{
+		set_nest_node_ses(ses, arg1, "%s", root->list[min]->arg1);
+
+		return;
+	}
+
+	max = get_number(ses, arg3);
+
+	if (max < 0)
+	{
+		max = root->used + max;
+	}
+	max = URANGE(0, max, root->used - 1);
+
+	if (min > max)
+	{
+		show_error(ses, LIST_COMMAND, "#ERROR: #HISTORY GET {%s} {%d} {%d} LOWER BOUND EXCEEDS UPPER BOUND.", arg1, min, max);
+
+		return;
+	}
+
+	cnt = 0;
+
+	set_nest_node_ses(ses, arg1, "");
+
+	while (min <= max)
+	{
+		sprintf(arg2, "%s[%d]", arg1, ++cnt);
+
+		set_nest_node_ses(ses, arg2, "%s", root->list[min++]->arg1);
+	}
+
+	show_message(ses, LIST_COMMAND, "#HISTORY GET: %d LINES SAVED TO {%s}.", cnt, arg1);
+
+	return;
 }
 
 DO_HISTORY(history_list)
 {
-	struct listroot *root;
-	int i, cnt = 1;
-
-	root = ses->list[LIST_HISTORY];
-
-	for (i = 0 ; i < root->used ; i++)
+	if (*arg1 == 0)
 	{
-		tintin_printf2(ses, "%6d - %s", cnt++, root->list[i]->left);
+		show_list(ses->list[LIST_HISTORY], 0);
+	}
+	else
+	{
+		show_node_with_wild(ses, arg1, ses->list[LIST_HISTORY]);
 	}
 	return;
 }
@@ -163,11 +220,11 @@ DO_HISTORY(history_read)
 	FILE *file;
 	char *cptr, buffer[BUFFER_SIZE];
 
-	file = fopen(arg, "r");
+	file = fopen(arg1, "r");
 
 	if (file == NULL)
 	{
-		show_message(ses, LIST_HISTORY, "#HISTORY: COULDN'T OPEN FILE {%s} TO READ.", arg);
+		show_message(ses, LIST_HISTORY, "#HISTORY: COULDN'T OPEN FILE {%s} TO READ.", arg1);
 		return;
 	}
 
@@ -183,21 +240,17 @@ DO_HISTORY(history_read)
 
 			if (*buffer)
 			{
-				insert_node_list(ses->list[LIST_HISTORY], buffer, "", "");
+				create_node_list(ses->list[LIST_HISTORY], buffer, "", "", "");
 			}
 		}
 	}
-	insert_node_list(ses->list[LIST_HISTORY], "", "", "");
+	create_node_list(ses->list[LIST_HISTORY], "", "", "", "");
 
 	fclose(file);
 
 	if (ses->list[LIST_HISTORY]->used > gtd->history_size) 
 	{
-		char buf[BUFFER_SIZE];
-
-		sprintf(buf, "{HISTORY SIZE} {%d}", UMIN(ses->list[LIST_HISTORY]->used, 9999));
-
-		do_configure(gts, buf);
+		execute(gts, "#CONFIG {HISTORY SIZE} {%d}", UMIN(ses->list[LIST_HISTORY]->used, 9999));
 	}
 
 	return;
@@ -205,13 +258,13 @@ DO_HISTORY(history_read)
 
 DO_HISTORY(history_size)
 {
-	if (atoi(arg) < 1 || atoi(arg) > 100000)
+	if (atoi(arg1) < 1 || atoi(arg1) > 100000)
 	{
-		tintin_printf(ses, "#HISTORY SIZE: PROVIDE A NUMBER BETWEEN 1 and 100,000");
+		show_error(ses, LIST_COMMAND, "#ERROR: #HISTORY SIZE: PROVIDE A NUMBER BETWEEN 1 and 100,000");
 	}
 	else
 	{
-		gtd->history_size = atoi(arg);
+		gtd->history_size = atoi(arg1);
 	}
 	return;
 }
@@ -222,18 +275,18 @@ DO_HISTORY(history_write)
 	FILE *file;
 	int i;
 
-	file = fopen(arg, "w");
+	file = fopen(arg1, "w");
 
 	if (file == NULL)
 	{
-		tintin_printf2(ses, "#HISTORY: COULDN'T OPEN FILE {%s} TO WRITE.", arg);
+		tintin_printf2(ses, "#HISTORY: COULDN'T OPEN FILE {%s} TO WRITE.", arg1);
 
 		return;
 	}
 
 	for (i = 0 ; i < root->used ; i++)
 	{
-		fprintf(file, "%s\n", root->list[i]->left);
+		fprintf(file, "%s\n", root->list[i]->arg1);
 	}
 
 	fclose(file);
