@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2005 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,14 +13,12 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
-*******************************************************************************/
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
+******************************************************************************/
 
 /******************************************************************************
-*   file: list.c - pseudo array support                                       *
-*                                                                             *
 *               (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                  *
 *                                                                             *
 *                     coded by Igor van den Hoven 2004                        *
@@ -33,19 +30,30 @@
 
 DO_COMMAND(do_list)
 {
-	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE];
-	struct listroot *root;
 	struct listnode *node;
-	int cnt;
-
-	root = ses->list[LIST_VARIABLE];
+	int index, cnt;
 
 	arg = sub_arg_in_braces(ses, arg, arg1, GET_NST, SUB_VAR|SUB_FUN);
 	arg = sub_arg_in_braces(ses, arg, arg2, GET_ONE, SUB_VAR|SUB_FUN);
 
-	if (*arg1 == 0 || *arg2 == 0)
+	if (*arg1 == 0)
 	{
-		show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} {ADD|CLE|CRE|DEL|FIN|GET|INS|SET|SIM|SIZ|SOR} {argument}");
+		info:
+
+		tintin_header(ses, " LIST OPTIONS ");
+
+		for (index = 0 ; *array_table[index].fun ; index++)
+		{
+			if (array_table[index].desc && *array_table[index].name)
+			{
+				tintin_printf2(ses, "  [%-24s] %s", array_table[index].name, array_table[index].desc);
+			}
+		}
+		tintin_header(ses, "");
+	}
+	else if (*arg2 == 0)
+	{
+		show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} {option} {argument}");
 	}
 	else
 	{
@@ -59,15 +67,15 @@ DO_COMMAND(do_list)
 
 		if (*array_table[cnt].name == 0)
 		{
-			return do_list(ses, "");
+			goto info;
 		}
 		else
 		{
-			if ((node = search_nest_node(root, arg1)) == NULL)
+			if ((node = search_nest_node_ses(ses, arg1)) == NULL)
 			{
-				node = set_nest_node(root, arg1, "");
+				node = set_nest_node_ses(ses, arg1, "");
 			}
-			array_table[cnt].array(ses, node, arg, arg1);
+			array_table[cnt].fun(ses, node, arg, arg1);
 		}
 	}
 	return ses;
@@ -97,13 +105,16 @@ DO_ARRAY(array_add)
 
 			set_nest_node(list->root, ntos(index++), "%s", arg2);
 
-//			insert_node_list(list->root, ntos(index++), arg2, "");
-
 			if (*str == COMMAND_SEPARATOR)
 			{
 				str++;
 			}
 		}
+
+		if (*arg == COMMAND_SEPARATOR)
+		{
+			arg++;
+		}
 	}
 	return ses;
 }
@@ -117,11 +128,30 @@ DO_ARRAY(array_clear)
 		list->root = NULL;
 	}
 
-	set_nest_node(ses->list[LIST_VARIABLE], var, "");
+	set_nest_node_ses(ses, var, "");
 
 	return ses;
 }
 
+DO_ARRAY(array_collapse)
+{
+	int index;
+
+	if (list->root)
+	{
+		str_cpy(&list->arg2, "");
+
+		for (index = 0 ; index < list->root->used ; index++)
+		{
+			str_cat(&list->arg2, list->root->list[index]->arg2);
+		}
+		free_list(list->root);
+
+		list->root = NULL;
+	}
+	return ses;
+}
+
 DO_ARRAY(array_create)
 {
 	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], buf[BUFFER_SIZE], *str;
@@ -151,8 +181,6 @@ DO_ARRAY(array_create)
 
 			set_nest_node(list->root, ntos(index++), "%s", arg2);
 
-//			insert_node_list(list->root, ntos(index++), arg2, "");
-
 			if (*str == COMMAND_SEPARATOR)
 			{
 				str++;
@@ -185,14 +213,14 @@ DO_ARRAY(array_delete)
 
 			if (atoi(arg1) == 0 || index == -1)
 			{
-				return show_error(ses, LIST_VARIABLE, "#LIST DEL: Invalid index: %s", arg1);
+				show_error(ses, LIST_VARIABLE, "#LIST DEL: Invalid index: %s", arg1);
 
 				return ses;
 			}
 
 			for (cnt = index + 1 ; cnt < list->root->used ; cnt++)
 			{
-				list->root->list[cnt]->left = refstring(list->root->list[cnt]->left, "%d", cnt);
+				str_cpy_printf(&list->root->list[cnt]->arg1, "%d", cnt);
 			}
 
 			delete_index_list(list->root, index);
@@ -205,6 +233,61 @@ DO_ARRAY(array_delete)
 	return ses;
 }
 
+DO_ARRAY(array_explode)
+{
+	char buf[BUFFER_SIZE], tmp[BUFFER_SIZE], *pti;
+	int index = 1;
+
+	if (list->root)
+	{
+		array_collapse(ses, list, "", "");
+	}
+
+	list->root = init_list(ses, LIST_VARIABLE, LIST_SIZE);
+
+	pti = list->arg2;
+
+	while (*pti)
+	{
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
+		{
+			pti += sprintf(tmp, "%.*s", get_euc_size(ses, pti), pti);
+		}
+		else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(pti))
+		{
+			pti += sprintf(tmp, "%.*s", get_utf8_size(pti), pti);
+		}
+		else
+		{
+			pti += sprintf(tmp, "%c", *pti);
+		}
+
+		set_nest_node(list->root, ntos(index++), "%s", tmp);
+	}
+	sub_arg_in_braces(ses, arg, buf, GET_ALL, SUB_VAR|SUB_FUN);
+
+	pti = buf;
+
+	while (*pti)
+	{
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
+		{
+			pti += sprintf(tmp, "%.*s", get_euc_size(ses, pti), pti);
+		}
+		else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(pti))
+		{
+			pti += sprintf(tmp, "%.*s", get_utf8_size(pti), pti);
+		}
+		else
+		{
+			pti += sprintf(tmp, "%c", *pti);
+		}
+
+		set_nest_node(list->root, ntos(index++), "%s", tmp);
+	}
+	return ses;
+}
+
 DO_ARRAY(array_find)
 {
 	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE];
@@ -215,31 +298,33 @@ DO_ARRAY(array_find)
 
 	if (*arg2 == 0)
 	{
-		return show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} FIND {string} {variable}");
+		show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} FIND {string} {variable}");
+		
+		return ses;
 	}
 
 	if (list->root)
 	{
 		for (index = 0 ; index < list->root->used ; index++)
 		{
-			if (match(ses, list->root->list[index]->right, arg1, SUB_NONE))
+			if (match(ses, list->root->list[index]->arg2, arg1, SUB_NONE))
 			{
 				break;
 			}
 		}
 		if (index < list->root->used)
 		{
-			set_nest_node(ses->list[LIST_VARIABLE], arg2, "%d", index + 1);
+			set_nest_node_ses(ses, arg2, "%d", index + 1);
 		}
 		else
 		{
-			set_nest_node(ses->list[LIST_VARIABLE], arg2, "0");
+			set_nest_node_ses(ses, arg2, "0");
 		}
 		return ses;
 	}
 	else
 	{
-		set_nest_node(ses->list[LIST_VARIABLE], arg2, "0");
+		set_nest_node_ses(ses, arg2, "0");
 	}
 
 	return ses;
@@ -255,7 +340,9 @@ DO_ARRAY(array_get)
 
 	if (*arg2 == 0)
 	{
-		return show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} GET {index} {variable}");
+		show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} GET {index} {variable}");
+		
+		return ses;
 	}
 
 	if (list->root)
@@ -264,17 +351,17 @@ DO_ARRAY(array_get)
 
 		if (atoi(arg1) == 0 || index == -1)
 		{
-			set_nest_node(ses->list[LIST_VARIABLE], arg2, "0");
+			set_nest_node_ses(ses, arg2, "0");
 		}
 		else
 		{
-			set_nest_node(ses->list[LIST_VARIABLE], arg2, "%s", list->root->list[index]->right);
+			set_nest_node_ses(ses, arg2, "%s", list->root->list[index]->arg2);
 		}
 		return ses;
 	}
 	else
 	{
-		set_nest_node(ses->list[LIST_VARIABLE], arg2, "0");
+		set_nest_node_ses(ses, arg2, "0");
 	}
 
 	return ses;
@@ -297,7 +384,9 @@ DO_ARRAY(array_insert)
 
 	if (atoi(arg1) == 0)
 	{
-		return show_error(ses, LIST_VARIABLE, "#LIST INS: Invalid index: %s", arg1);
+		show_error(ses, LIST_VARIABLE, "#LIST INS: Invalid index: %s", arg1);
+		
+		return ses;
 	}
 
 	if (index == -1 || atoi(arg1) < 0)
@@ -307,48 +396,101 @@ DO_ARRAY(array_insert)
 
 	for (cnt = index ; cnt < list->root->used ; cnt++)
 	{
-		list->root->list[cnt]->left = refstring(list->root->list[cnt]->left, "%d", cnt + 2);
+		str_cpy_printf(&list->root->list[cnt]->arg1, "%d", cnt + 2);
 	}
 
 	set_nest_node(list->root, ntos(index + 1), "%s", arg2);
 
-//	insert_node_list(list->root, ntos(index + 1), arg2, "");
+	return ses;
+}
 
+DO_ARRAY(array_order)
+{
+	int cnt;
+	char **buffer;
+
+	array_add(ses, list, arg, var);
+
+	buffer = malloc(list->root->used * sizeof(char *));
+
+	for (cnt = 0 ; cnt < list->root->used ; cnt++)
+	{
+		buffer[cnt] = list->root->list[cnt]->arg2;
+	}
+
+	quadsort(buffer, list->root->used, sizeof(char *), cmp_num);
+
+	for (cnt = 0 ; cnt < list->root->used ; cnt++)
+	{
+		list->root->list[cnt]->arg2 = buffer[cnt];
+	}
+
+	free(buffer);
+
+	return ses;
+}
+
+DO_ARRAY(array_reverse)
+{
+	char *swap;
+	int cnt, rev;
+
+	array_add(ses, list, arg, var);
+
+	for (cnt = 0 ; cnt < list->root->used / 2 ; cnt++)
+	{
+		rev = list->root->used - 1 - cnt;
+
+		swap = list->root->list[cnt]->arg2;
+		list->root->list[cnt]->arg2 = list->root->list[rev]->arg2;
+		list->root->list[rev]->arg2 = swap;
+	}
 	return ses;
 }
 
 DO_ARRAY(array_simplify)
 {
-	char arg1[BUFFER_SIZE], tmp[BUFFER_SIZE];
+	char arg1[BUFFER_SIZE], *str;
 	int index;
 
 	arg = sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
-
+/*
 	if (*arg1 == 0)
 	{
-		return show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} SIMPLIFY {variable}");
+		show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} SIMPLIFY {variable}");
+		
+		return ses;
 	}
-
+*/
 	if (list->root)
 	{
 		for (index = 0 ; index < list->root->used ; index++)
 		{
 			if (index == 0)
 			{
-				strcpy(tmp, list->root->list[index]->right);
+				str = str_dup(list->root->list[index]->arg2);
 			}
 			else
 			{
-				cat_sprintf(tmp, ";%s", list->root->list[index]->right);
+				str_cat_printf(&str, ";%s", list->root->list[index]->arg2);
 			}
 		}
-		set_nest_node(ses->list[LIST_VARIABLE], arg1, "%s", tmp);
+		if (*arg1 == 0)
+		{
+			set_nest_node_ses(ses, list->arg1, "%s", str);
+		}
+		else
+		{
+			set_nest_node_ses(ses, arg1, "%s", str);
+		}
+
+		str_free(str);
 
 		return ses;
 	}
 	else
 	{
-		show_error(ses, LIST_VARIABLE, "#LIST SIMPLIFY: {%s} is not a list.", arg1);
+		show_error(ses, LIST_VARIABLE, "#LIST SIMPLIFY: {%s} is not a list.", list->arg1);
 	}
 
 	return ses;
@@ -362,16 +504,18 @@ DO_ARRAY(array_size)
 
 	if (*arg1 == 0)
 	{
-		return show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} SIZE {variable}");
+		show_error(ses, LIST_VARIABLE, "#SYNTAX: #LIST {variable} SIZE {variable}");
+		
+		return ses;
 	}
 
 	if (list->root)
 	{
-		set_nest_node(ses->list[LIST_VARIABLE], arg1, "%d", list->root->used);
+		set_nest_node_ses(ses, arg1, "%d", list->root->used);
 	}
 	else
 	{
-		set_nest_node(ses->list[LIST_VARIABLE], arg1, "0");
+		set_nest_node_ses(ses, arg1, "0");
 	}
 	return ses;
 }
@@ -389,10 +533,14 @@ DO_ARRAY(array_set)
 
 		if (atoi(arg1) == 0 || index == -1)
 		{
-			return show_error(ses, LIST_VARIABLE, "#LIST SET: Invalid index: %s", arg1);
+			show_error(ses, LIST_VARIABLE, "#LIST SET: Invalid index: %s", arg1);
+			
+			return ses;
 		}
 
-		RESTRING(list->root->list[index]->right, arg2);
+//		set_nest_node(list->root, ntos(index + 1), "%s", arg2);
+
+		str_cpy(&list->root->list[index]->arg2, arg2);
 
 		return ses;
 	}
@@ -402,55 +550,47 @@ DO_ARRAY(array_set)
 	return ses;
 }
 
+DO_ARRAY(array_shuffle)
+{
+	char *swap;
+	int cnt, rnd;
+
+	array_add(ses, list, arg, var);
+
+	for (cnt = 0 ; cnt < list->root->used ; cnt++)
+	{
+		rnd = generate_rand(ses) % list->root->used;
+
+		swap = list->root->list[cnt]->arg2;
+		list->root->list[cnt]->arg2 = list->root->list[rnd]->arg2;
+		list->root->list[rnd]->arg2 = swap;
+	}
+	return ses;
+}
+
 DO_ARRAY(array_sort)
 {
-	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], arg3[BUFFER_SIZE], *str;
 	int cnt;
+	char **buffer;
 
-	if (!list->root)
+	array_add(ses, list, arg, var);
+
+	buffer = malloc(list->root->used * sizeof(char *));
+
+	for (cnt = 0 ; cnt < list->root->used ; cnt++)
 	{
-		list->root = init_list(ses, LIST_VARIABLE, LIST_SIZE);
+		buffer[cnt] = list->root->list[cnt]->arg2;
 	}
 
-	while (*arg)
+	quadsort(buffer, list->root->used, sizeof(char *), cmp_str);
+
+	for (cnt = 0 ; cnt < list->root->used ; cnt++)
 	{
-		arg = sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
-
-		str = arg1;
-
-		while (*str)
-		{
-			str = get_arg_in_braces(ses, str, arg2, GET_ALL);
-
-			for (cnt = 0 ; cnt < list->root->used ; cnt++)
-			{
-				if (strcmp(arg2, list->root->list[cnt]->right) <= 0)
-				{
-					break;
-				}
-			}
-
-			if (cnt == list->root->used)
-			{
-				sprintf(arg3, "{%d} {%s}", -1, arg2);
-			}
-			else
-			{
-				sprintf(arg3, "{%d} {%s}", cnt + 1, arg2);
-			}
-
-			array_insert(ses, list, arg3, var);
-
-			if (*str == COMMAND_SEPARATOR)
-			{
-				str++;
-			}
-		}
-		if (*arg == COMMAND_SEPARATOR)
-		{
-			arg++;
-		}
+		list->root->list[cnt]->arg2 = buffer[cnt];
 	}
+
+	free(buffer);
+
 	return ses;
 }
 
@@ -472,24 +612,13 @@ DO_ARRAY(array_tokenize)
 
 	while (buf[i] != 0)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && buf[i] & 128 && buf[i+1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, &buf[i]))
 		{
-			i += sprintf(tmp, "%c%c", buf[i], buf[i+1]);
+			i += sprintf(tmp, "%.*s", get_euc_size(ses, &buf[i]), &buf[i]);
 		}
-		else if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (buf[i] & 192) == 192 && buf[i+1] != 0)
+		else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(&buf[i]))
 		{
-			if ((buf[i] & 240) == 240 && buf[i+2] != 0 && buf[i+3] != 0)
-			{
-				i += sprintf(tmp, "%c%c%c%c", buf[i], buf[i+1], buf[i+2], buf[i+3]);
-			}
-			else if ((buf[i] & 224) == 224 && buf[i+2] != 0)
-			{
-				i += sprintf(tmp, "%c%c%c", buf[i], buf[i+1], buf[i+2]);
-			}
-			else
-			{
-				i += sprintf(tmp, "%c%c", buf[i], buf[i+1]);
-			}
+			i += sprintf(tmp, "%.*s", get_utf8_size(&buf[i]), &buf[i]);
 		}
 		else
 		{
