@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2020 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -15,12 +14,11 @@
 *   GNU General Public License for more details.                              *
 *                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
 ******************************************************************************/
 
 /******************************************************************************
-*                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
+*                               T I N T I N + +                               *
 *                                                                             *
 *                      coded by Igor van den Hoven 2004                       *
 ******************************************************************************/
@@ -29,7 +27,8 @@
 
 #define EXP_VARIABLE         0
 #define EXP_STRING           1
-#define EXP_OPERATOR         2
+#define EXP_BRACE            2
+#define EXP_OPERATOR         3
 
 #define EXP_PR_DICE          0
 #define EXP_PR_INTMUL        1
@@ -55,25 +54,24 @@ int precision;
 
 DO_COMMAND(do_math)
 {
-	char left[BUFFER_SIZE], right[BUFFER_SIZE];
 	struct listnode *node;
-	double result;
+	long double result;
 
-	arg = sub_arg_in_braces(ses, arg, left, GET_NST, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_NST, SUB_VAR|SUB_FUN);
 
-	arg = get_arg_in_braces(ses, arg, right, GET_ALL);
+	arg = get_arg_in_braces(ses, arg, arg2, GET_ALL);
 
-	if (*left == 0 || *right == 0)
+	if (*arg1 == 0 || *arg2 == 0)
 	{
 		show_error(ses, LIST_VARIABLE, "#SYNTAX: #MATH {variable} {expression}.");
 	}
 	else
 	{
-		result = get_number(ses, right);
+		result = get_number(ses, arg2);
 
-		node = set_nest_node(ses->list[LIST_VARIABLE], left, "%.*f", precision, result);
+		node = set_nest_node_ses(ses, arg1, "%.*Lf", precision, result);
 
-		show_message(ses, LIST_VARIABLE, "#MATH: VARIABLE {%s} HAS BEEN SET TO {%s}.", left, node->right);
+		show_message(ses, LIST_VARIABLE, "#MATH: VARIABLE {%s} HAS BEEN SET TO {%s}.", arg1, node->arg2);
 	}
 
 	return ses;
@@ -84,9 +82,9 @@ int is_math(struct session *ses, char *str)
 	return mathexp_tokenize(ses, str, 0, 0);
 }
 
-double get_number(struct session *ses, char *str)
+long double get_number(struct session *ses, char *str)
 {
-	double val;
+	long double val;
 	char result[BUFFER_SIZE];
 
 	mathexp(ses, str, result, 0);
@@ -96,9 +94,41 @@ double get_number(struct session *ses, char *str)
 	return val;
 }
 
-double get_double(struct session *ses, char *str)
+unsigned long long get_integer(struct session *ses, char *str)
 {
-	double val;
+	unsigned long long val;
+	char result[BUFFER_SIZE];
+
+	mathexp(ses, str, result, 0);
+
+	val = tintou(result);
+
+	return val;
+}
+
+int get_ellipsis(struct listroot *root, char *name, int *min, int *max)
+{
+	unsigned long long range = get_integer(root->ses, name);
+
+	push_call("get_ellipsis(%p,%p,%p,%p)",root,name,min,max);
+
+
+	*min = (int) (HAS_BIT(range, 0x00000000FFFFFFFFLL));
+	*max = (int) (HAS_BIT(range, 0xFFFFFFFF00000000ULL) >> 32ULL);
+
+	*min = *min > 0 ? *min - 1 : root->used + *min;
+	*max = *max > 0 ? *max - 1 : root->used + *max;
+
+	*min = URANGE(0, *min, root->used - 1);
+	*max = URANGE(0, *max, root->used - 1);
+
+	pop_call();
+	return 1 + (*min < *max ? *max - *min : *min - *max);
+}
+
+long double get_double(struct session *ses, char *str)
+{
+	long double val;
 	char result[BUFFER_SIZE];
 
 	mathexp(ses, str, result, 1);
@@ -110,10 +140,12 @@ double get_double(struct session *ses, char *str)
 	
 void get_number_string(struct session *ses, char *str, char *result)
 {
-	sprintf(result, "%.*f", precision, get_number(ses, str));
+	long double val = get_number(ses, str);
+
+	sprintf(result, "%.*Lf", precision, val);
 }
 
-double mathswitch(struct session *ses, char *left, char *right)
+long double mathswitch(struct session *ses, char *left, char *right)
 {
 	char shift[BUFFER_SIZE];
 
@@ -167,7 +199,8 @@ void mathexp(struct session *ses, char *str, char *result, int seed)
 	if (badnumber && debug)                                 \
 	{                                                       \
 		badnumber = 0;                                  \
-		show_debug(ses, LIST_VARIABLE, "#MATH EXP: INVALID NUMBER %s.", buf3); \
+		precision = 0;                                  \
+		show_debug(ses, LIST_VARIABLE, "#MATH EXP {%s}: INVALID NUMBER %s.", str, buf3); \
 	}                                                       \
 	*pta = 0;                                               \
 	sprintf(buf1, "%02d", level);                           \
@@ -176,6 +209,7 @@ void mathexp(struct session *ses, char *str, char *result, int seed)
 	status = newstatus;                                     \
 	pta = buf3;                                             \
 	point = -1;                                             \
+	metric = 0;                                             \
 }
 
 void add_math_node(char *arg1, char *arg2, char *arg3)
@@ -205,9 +239,11 @@ void del_math_node(struct link_data *node)
 int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 {
 	char buf1[BUFFER_SIZE], buf2[BUFFER_SIZE], buf3[STRING_SIZE], *pti, *pta;
-	int level, status, point, badnumber;
+	int level, status, point, badnumber, metric, nest;
 
+	nest      = 0;
 	level     = 0;
+	metric    = 0;
 	point     = -1;
 	status    = EXP_VARIABLE;
 	precision = seed;
@@ -240,6 +276,16 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 					case '9':
 						*pta++ = *pti++;
 
+						if (metric)
+						{
+							badnumber = 1;
+
+							if (debug == 0)
+							{
+								return FALSE;
+							}
+						}
+
 						if (point >= 0)
 						{
 							point++;
@@ -252,8 +298,53 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						break;
 
 					case '!':
+						if (pta != buf3)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						}
+						else
+						{
+							add_math_node(ntos(level), ntos(EXP_PR_VAR), "0");
+							add_math_node(ntos(level), ntos(EXP_PR_LOGCOMP), "==");
+
+							*pta++ = *pti++;
+
+							pta = buf3;
+						}
+						break;
+
 					case '~':
+						if (pta != buf3)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						}
+						else
+						{
+							add_math_node(ntos(level), ntos(EXP_PR_VAR), "-1");
+							add_math_node(ntos(level), ntos(EXP_PR_INTADD), "-");
+
+							*pta++ = *pti++;
+
+							pta = buf3;
+						}
+						break;
+
 					case '+':
+						if (pta != buf3)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						}
+						else
+						{
+							add_math_node(ntos(level), ntos(EXP_PR_VAR), "1");
+							add_math_node(ntos(level), ntos(EXP_PR_INTMUL), "*");
+
+							*pta++ = *pti++;
+
+							pta = buf3;
+						}
+						break;
+
 					case '-':
 						if (pta != buf3)
 						{
@@ -261,10 +352,29 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						}
 						else
 						{
+							add_math_node(ntos(level), ntos(EXP_PR_VAR), "-1");
+							add_math_node(ntos(level), ntos(EXP_PR_INTMUL), "*");
+
 							*pta++ = *pti++;
+
+							pta = buf3;
 						}
 						break;
 
+					case '{':
+						if (pta != buf3)
+						{
+							if (debug)
+							{
+								show_debug(ses, LIST_VARIABLE, "MATH EXP: { FOUND INSIDE A NUMBER");
+							}
+							return FALSE;
+						}
+						*pta++ = *pti++;
+						status = EXP_BRACE;
+						nest++;
+						break;
+
 					case '"':
 						if (pta != buf3)
 						{
@@ -275,6 +385,7 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 							return FALSE;
 						}
 						*pta++ = *pti++;
+						nest++;
 						status = EXP_STRING;
 						break;
 
@@ -287,8 +398,11 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 							}
 							return FALSE;
 						}
-						*pta++ = *pti++;
-						MATH_NODE(FALSE, EXP_PR_LVL, EXP_VARIABLE);
+						else
+						{
+							*pta++ = *pti++;
+							MATH_NODE(FALSE, EXP_PR_LVL, EXP_VARIABLE);
+						}
 						level++;
 						break;
 
@@ -301,17 +415,24 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						break;
 						
 					case '.':
-						*pta++ = *pti++;
-						if (point >= 0)
+						if (pta != buf3 && pti[1] == '.')
 						{
-							if (debug)
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						}
+						else
+						{
+							*pta++ = *pti++;
+							if (point >= 0)
 							{
-								show_debug(ses, LIST_VARIABLE, "#MATH EXP: MORE THAN ONE POINT FOUND INSIDE A NUMBER");
+								if (debug)
+								{
+									show_debug(ses, LIST_VARIABLE, "#MATH EXP: MORE THAN ONE POINT FOUND INSIDE A NUMBER");
+								}
+								precision = 0;
+								return FALSE;
 							}
-							precision = 0;
-							return FALSE;
+							point++;
 						}
-						point++;
 						break;
 
 					case ' ':
@@ -319,8 +440,8 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						pti++;
 						break;
 
-					case ')':
 					case 'd':
+					case ')':
 					case '*':
 					case '/':
 					case '%':
@@ -330,6 +451,15 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 					case '^':
 					case '|':
 					case '=':
+						if (pti == str)
+						{
+							if (debug)
+							{
+								show_debug(ses, LIST_VARIABLE, "#MATH EXP: EXPRESSION STARTED WITH AN OPERATOR.");
+							}
+							return FALSE;
+						}
+
 						if (pta != buf3)
 						{
 							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
@@ -340,6 +470,98 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						}
 						break;
 
+					case 'K':
+					case 'M':
+//					case 'G':
+//					case 'T':
+//					case 'P':
+//					case 'E':
+//					case 'Z':
+//					case 'Y':
+						if (pta == buf3 || metric == 1)
+						{
+							badnumber = 1;
+
+							if (debug == 0)
+							{
+								return FALSE;
+							}
+							*pta++ = *pti++;
+							*pta = 0;
+						}
+						else if (badnumber == 0)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+
+							*pta++ = '*';
+							MATH_NODE(FALSE, EXP_PR_INTMUL, EXP_VARIABLE);
+
+							switch (*pti++)
+							{
+								case 'K': pta += sprintf(pta, "1000"); break;
+								case 'M': pta += sprintf(pta, "1000000"); break;
+								case 'G': pta += sprintf(pta, "1000000000"); break;
+								case 'T': pta += sprintf(pta, "1000000000000"); break;
+								case 'P': pta += sprintf(pta, "1000000000000000"); break;
+								case 'E': pta += sprintf(pta, "1000000000000000000"); break;
+								case 'Z': pta += sprintf(pta, "1000000000000000000000"); break;
+								case 'Y': pta += sprintf(pta, "1000000000000000000000000"); break;
+							}
+							metric = 1;
+						}
+						else
+						{
+							*pta++ = *pti++;
+						}
+						break;
+
+					case 'm':
+					case 'u':
+//					case 'n':
+//					case 'p':
+//					case 'f':
+//					case 'a':
+//					case 'z':
+//					case 'y':
+						if (pta == buf3 || metric == 1)
+						{
+							badnumber = 1;
+
+							if (debug == 0)
+							{
+								return FALSE;
+							}
+							*pta++ = *pti++;
+							*pta = 0;
+						}
+						else if (badnumber == 0)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+
+							*pta++ = '/';
+
+							MATH_NODE(FALSE, EXP_PR_INTMUL, EXP_VARIABLE);
+
+							switch (*pti++)
+							{
+								case 'm': pta += sprintf(pta, "1000"); break;
+								case 'u': pta += sprintf(pta, "1000000"); break;
+								case 'n': pta += sprintf(pta, "1000000000"); break;
+								case 'p': pta += sprintf(pta, "1000000000000"); break;
+								case 'f': pta += sprintf(pta, "1000000000000000"); break;
+								case 'a': pta += sprintf(pta, "1000000000000000000"); break;
+								case 'z': pta += sprintf(pta, "1000000000000000000000"); break;
+								case 'y': pta += sprintf(pta, "1000000000000000000000000"); break;
+							}
+							metric = 1;
+							precision = UMAX(precision, strlen(buf3) -1);
+						}
+						else
+						{
+							*pta++ = *pti++;
+						}
+						break;
+
 					default:
 						*pta++ = *pti++;
 						*pta = 0;
@@ -359,7 +581,34 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 				{
 					case '"':
 						*pta++ = *pti++;
-						MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						nest--;
+						if (nest == 0)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						}
+						break;
+
+					default:
+						*pta++ = *pti++;
+						break;
+				}
+				break;
+
+			case EXP_BRACE:
+				switch (*pti)
+				{
+					case '{':
+						*pta++ = *pti++;
+						nest++;
+						break;
+
+					case '}':
+						*pta++ = *pti++;
+						nest--;
+						if (nest == 0)
+						{
+							MATH_NODE(FALSE, EXP_PR_VAR, EXP_OPERATOR);
+						}
 						break;
 
 					default:
@@ -375,6 +624,25 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						pti++;
 						break;
 
+					case '.':
+						if (pti[1] == '.')
+						{
+							*pta++ = *pti++;
+							*pta++ = *pti++;
+
+							MATH_NODE(FALSE, EXP_PR_LOGCOMP, EXP_VARIABLE);
+							break;
+						}
+						else
+						{
+							if (debug)
+							{
+								show_debug(ses, LIST_VARIABLE, "#MATH EXP: UNKNOWN OPERATOR: %c%c", pti[0], pti[1]);
+							}
+							return FALSE;
+						}
+						break;
+
 					case ')':
 						*pta++ = *pti++;
 						level--;
@@ -526,6 +794,10 @@ int mathexp_tokenize(struct session *ses, char *str, int seed, int debug)
 						{
 							case '=':
 								*pta++ = *pti++;
+								if (*pti == '=')
+								{
+									*pta++ = *pti++;
+								}
 								MATH_NODE(FALSE, EXP_PR_LOGCOMP, EXP_VARIABLE);
 								break;
 
@@ -622,10 +894,22 @@ void mathexp_level(struct session *ses, struct link_data *node)
 void mathexp_compute(struct session *ses, struct link_data *node)
 {
 	char temp[BUFFER_SIZE];
-	double value;
+	int integer64 = 0;
+	long double value = 0;
+	unsigned long long min = 0, max = 0;
+	unsigned long long value64 = 0;
 
 	switch (node->str3[0])
 	{
+		case '.':
+			integer64 = 1;
+
+                        SET_BIT(max, (unsigned int) tintoi(node->next->str3));
+                        max = max << 32ULL;
+                        SET_BIT(min, (unsigned int) tintoi(node->prev->str3));
+                        value64 = max | min;
+                        break;
+
 		case 'd':
 			if (tintoi(node->next->str3) <= 0)
 			{
@@ -634,7 +918,7 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 			}
 			else
 			{
-				value = tindice(node->prev->str3, node->next->str3);
+				value = tindice(ses, node->prev->str3, node->next->str3);
 			}
 			break;
 		case '*':
@@ -665,7 +949,9 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 					if (tintoi(node->next->str3) == 0)
 					{
 						show_debug(ses, LIST_VARIABLE, "#MATH ERROR: DIVISION ZERO.");
-						value = tintoi(node->prev->str3);
+						value = 0;
+						precision = 0;
+//						value = tintoi(node->prev->str3);
 					}
 					else
 					{
@@ -689,9 +975,9 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 			}
 			else
 			{
-				// Can't perform modulo on doubles
+				integer64 = 1;
 
-				value = (long long) tintoi(node->prev->str3) % (long long) tintoi(node->next->str3);
+				value64 = tintou(node->prev->str3) % tintou(node->next->str3);
 			}
 			break;
 		case '+':
@@ -700,15 +986,18 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 		case '-':
 			value = tintoi(node->prev->str3) - tintoi(node->next->str3);
 			break;
+
 		case '<':
 			switch (node->str3[1])
 			{
 				case '=':
 					value = tincmp(node->prev->str3, node->next->str3) <= 0;
 					break;
+
 				case '<':
 					value = (long long) tintoi(node->prev->str3) << (long long) tintoi(node->next->str3);
 					break;
+
 				default:
 					value = tincmp(node->prev->str3, node->next->str3) < 0;
 					break;
@@ -720,9 +1009,11 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 				case '=':
 					value = tincmp(node->prev->str3, node->next->str3) >= 0;
 					break;
+
 				case '>':
 					value = (long long) tintoi(node->prev->str3) >> (long long) tintoi(node->next->str3);
 					break;
+
 				default:
 					value = tincmp(node->prev->str3, node->next->str3) > 0;
 					break;
@@ -735,6 +1026,7 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 				case '&':
 					value = tintoi(node->prev->str3) && tintoi(node->next->str3);
 					break;
+
 				default:
 					value = (long long) tintoi(node->prev->str3) & (long long) tintoi(node->next->str3);
 					break;
@@ -765,11 +1057,27 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 			}
 			break;
 		case '=':
-			value = (tineval(ses, node->prev->str3, node->next->str3) != 0);
+			if (node->str3[1] == '=' && node->str3[2] == '=')
+			{
+				
+				value = tincmp(node->prev->str3, node->next->str3) == 0;
+			}
+			else
+			{
+				value = tineval(ses, node->prev->str3, node->next->str3) != 0;
+			}
 			break;
 		case '!':
-			value = (tineval(ses, node->prev->str3, node->next->str3) == 0);
+			if (node->str3[1] == '=' && node->str3[2] == '=')
+			{
+				value = tincmp(node->prev->str3, node->next->str3) != 0;
+			}
+			else
+			{
+				value = tineval(ses, node->prev->str3, node->next->str3) == 0;
+			}
 			break;
+
 		default:
 			show_debug(ses, LIST_VARIABLE, "#COMPUTE EXP: UNKNOWN OPERATOR: %c%c", node->str3[0], node->str3[1]);
 			value = 0;
@@ -793,7 +1101,14 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 	free(node->str2);
 	node->str2 = strdup(temp);
 
-	sprintf(temp, "%lf", value);
+	if (integer64)
+	{
+		sprintf(temp, "%llu", value64);
+	}
+	else
+	{
+		sprintf(temp, "%.*Lf", precision, value);
+	}
 	free(node->str3);
 	node->str3 = strdup(temp);
 }
@@ -802,10 +1117,10 @@ void mathexp_compute(struct session *ses, struct link_data *node)
 	Keep synced with is_number()
 */
 
-double tintoi(char *str)
+long double tintoi(char *str)
 {
 	char *ptr = str;
-	double values[5] = {0, 0, 0, 0, 0}, m = 1;
+	long double values[5] = {0, 0, 0, 0, 0}, m = 1;
 	int i = 1, d = 0;
 
 	if (*ptr == 0)
@@ -918,75 +1233,174 @@ double tintoi(char *str)
 		default:
 			return (values[0] + values[1] + values[2] + values[3] + values[4]);
 	}
+}
 
-	switch (str[0])
+unsigned long long tintou(char *str)
+{
+	char *ptr = str;
+	unsigned long long value = 0, m = 1;
+
+	if (*ptr == 0)
 	{
-		case '!':
-			return !atof(&str[1]);
-
-		case '~':
-			return (double) ~atoll(&str[1]);
-
-		case '+':
-			return +atof(&str[1]);
-
-		case '-':
-			return -atof(&str[1]);
-
-		default:
-			return atof(str);
+		return 0;
 	}
 
-	while (*ptr)
+	switch (*ptr)
 	{
-		if (!isdigit((int) *ptr))
+		case '!':
+		case '~':
+		case '+':
+		case '-':
+			ptr++;
+			break;
+	}
+
+	ptr = str + strlen(str);
+
+	while (ptr != str)
+	{
+		ptr--;
+
+		switch (*ptr)
 		{
-			if (*ptr != '.')
-			{
+			case '0':
+			case '1':
+			case '2':
+			case '3':
+			case '4':
+			case '5':
+			case '6':
+			case '7':
+			case '8':
+			case '9':
+				value += (*ptr - '0') * m;
+				m *= 10;
+				break;
+
+			case '.':
+				*ptr = 0;
+				break;
+
+			case ':':
 				return 0;
-			}
-			ptr++;
+				break;
 
-			while (*ptr)
-			{
-				if (!isdigit((int) *ptr))
+			case '!':
+			case '~':
+			case '+':
+			case '-':
+				if (ptr == str)
 				{
-					return 0;
+					break;
 				}
-				ptr++;
-			}
-		}
-		else
-		{
-			ptr++;
+				return 0;
+
+			default:
+				return 0;
 		}
 	}
 
 	switch (str[0])
 	{
 		case '!':
-			return !atof(&str[1]);
+			return !value;
 
 		case '~':
-			return (double) ~atoll(&str[1]);
+			return ~value;
 
 		case '+':
-			return +atof(&str[1]);
+			return +value;
 
 		case '-':
-			return -atof(&str[1]);
+			return -value;
 
 		default:
-			return atof(str);
+			return value;
 	}
 }
 
-double tincmp(char *left, char *right)
+/*
+	Keep synched with tintoi() and is_math()
+*/
+
+int is_number(char *str)
 {
-	double left_val, right_val;
+	char *ptr = str;
+	int i = 1, d = 0, valid = 0;
+
+	if (*ptr == 0)
+	{
+		return FALSE;
+	}
+
+	ptr = str + strlen(str);
+
+	while (TRUE)
+	{
+		ptr--;
+
+		switch (*ptr)
+		{
+			case '0':
+			case '1':
+			case '2':
+			case '3':
+			case '4':
+			case '5':
+			case '6':
+			case '7':
+			case '8':
+			case '9':
+				valid = TRUE;
+				break;
+
+			case '.':
+				if (d)
+				{
+					return FALSE;
+				}
+				d = 1;
+				valid = FALSE;
+				break;
+
+			case ':':
+				if (i == 4)
+				{
+					return FALSE;
+				}
+				i++;
+				valid = FALSE;
+				break;
+
+			case '!':
+			case '~':
+			case '+':
+			case '-':
+				if (ptr != str)
+				{
+					return FALSE;
+				}
+				break;
+
+			default:
+				return FALSE;
+		}
+
+		if (ptr == str)
+		{
+			break;
+		}
+	}
+	return valid;
+}
+
+long double tincmp(char *left, char *right)
+{
+	long double left_val, right_val;
 
 	switch (left[0])
 	{
+		case '{':
 		case '"':
 			return strcmp(left, right);
 
@@ -998,12 +1412,18 @@ double tincmp(char *left, char *right)
 	}
 }
 
-double tineval(struct session *ses, char *left, char *right)
+long double tineval(struct session *ses, char *left, char *right)
 {
-	double left_val, right_val;
+	long double left_val, right_val;
 
 	switch (left[0])
 	{
+		case '{':
+			get_arg_in_braces(ses, left, left, GET_ALL);
+			get_arg_in_braces(ses, right, right, GET_ALL);
+
+			return match(ses, left, right, SUB_NONE);
+
 		case '"':
 			return match(ses, left, right, SUB_NONE);
 
@@ -1015,17 +1435,35 @@ double tineval(struct session *ses, char *left, char *right)
 	}
 }
 
-double tindice(char *left, char *right)
+long double tindice(struct session *ses, char *left, char *right)
 {
-	long long cnt, numdice, sizedice, sum;
+	unsigned long long cnt, numdice, sizedice, sum;
+	long double estimate;
 
-	numdice  = (long long) tintoi(left);
-	sizedice = (long long) tintoi(right);
+	numdice  = (unsigned long long) tintoi(left);
+	sizedice = (unsigned long long) tintoi(right);
+
+	if (sizedice == 0)
+	{
+		return 0;
+	}
+
+	if (numdice > 100)
+	{
+		estimate = numdice / 100.0;
+		numdice = 100;
+	}
+	else
+	{
+		estimate = 1;
+	}
 
 	for (cnt = sum = 0 ; cnt < numdice ; cnt++)
 	{
-		sum += rand() % sizedice + 1;
+		sum += generate_rand(ses) % sizedice + 1;
 	}
 
-	return (double) sum;
+	sum *= estimate;
+
+	return (long double) sum;
 }
