@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,10 +13,10 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
-*******************************************************************************/
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
+******************************************************************************/
 
 /******************************************************************************
 *                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
@@ -30,7 +29,77 @@
 
 DO_COMMAND(do_bell)
 {
-	printf("\007");
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg2, GET_ONE, SUB_VAR|SUB_FUN);
+
+	if (*arg1 == 0)
+	{
+		print_stdout("\007");
+	}
+	else if (is_abbrev(arg1, "FLASH"))
+	{
+		if (is_abbrev(arg2, "ON"))
+		{
+			print_stdout("\e[?1042h");
+		}
+		else if (is_abbrev(arg2, "OFF"))
+		{
+			print_stdout("\e[?1042l");
+		}
+		else
+		{
+			show_error(ses, LIST_COMMAND, "#SYNTAX: #BELL FLASH {ON|OFF}");
+		}
+	}
+	else if (is_abbrev(arg1, "FOCUS"))
+	{
+		if (is_abbrev(arg2, "ON"))
+		{
+			print_stdout("\e[?1043h");
+		}
+		else if (is_abbrev(arg2, "OFF"))
+		{
+			print_stdout("\e[?1043l");
+		}
+		else
+		{
+			show_error(ses, LIST_COMMAND, "#SYNTAX: #BELL FOCUS {ON|OFF}");
+		}
+	}
+	else if (is_abbrev(arg1, "MARGIN"))
+	{
+		if (is_abbrev(arg2, "ON"))
+		{
+			print_stdout("\e[?44h");
+		}
+		else if (is_abbrev(arg2, "OFF"))
+		{
+			print_stdout("\e[?44l");
+		}
+		else
+		{
+			show_error(ses, LIST_COMMAND, "#SYNTAX: #BELL MARGIN {ON|OFF}");
+		}
+	}
+	else if (is_abbrev(arg1, "RING"))
+	{
+		print_stdout("\007");
+	}
+	else if (is_abbrev(arg1, "VOLUME"))
+	{
+		if (is_math(ses, arg2))
+		{
+			print_stdout("\e[ %dt", (int) get_number(ses, arg2));
+		}
+		else
+		{
+			show_error(ses, LIST_COMMAND, "#SYNTAX: #BELL VOLUME {1-8}");
+		}
+	}
+	else
+	{
+		show_error(ses, LIST_COMMAND, "#SYNTAX: #BELL {FLASH|FOCUS|MARGIN|RING|VOLUME} {ARGUMENT}");
+	}
 
 	return ses;
 }
@@ -38,7 +107,7 @@ DO_COMMAND(do_bell)
 
 DO_COMMAND(do_commands)
 {
-	char buf[BUFFER_SIZE] = { 0 }, add[BUFFER_SIZE];
+	char buf[BUFFER_SIZE] = { 0 };
 	int cmd;
 
 	tintin_header(ses, " %s ", "COMMANDS");
@@ -49,20 +118,25 @@ DO_COMMAND(do_commands)
 		{
 			continue;
 		}
-		if ((int) strlen(buf) + 20 > ses->cols)
+
+		if (strip_vt102_strlen(ses, buf) + 20 > gtd->screen->cols)
 		{
 			tintin_puts2(ses, buf);
 			buf[0] = 0;
 		}
-		sprintf(add, "%20s", command_table[cmd].name);
-		strcat(buf, add);
+		if (command_table[cmd].type == TOKEN_TYPE_COMMAND)
+		{
+			cat_sprintf(buf, "%s%20s", COLOR_COMMAND, command_table[cmd].name);
+		}
+		else
+		{
+			cat_sprintf(buf, "%s%20s", COLOR_STATEMENT, command_table[cmd].name);
+		}
 	}
 	if (buf[0])
 	{
 		tintin_puts2(ses, buf);
 	}
-	tintin_header(ses, "");
-
 	return ses;
 }
 
@@ -88,8 +162,8 @@ DO_COMMAND(do_echo)
 
 	if (*arg == DEFAULT_OPEN)
 	{
-		arg = get_arg_in_braces(ses, arg, left, TRUE);
-		      get_arg_in_braces(ses, arg, temp, TRUE);
+		arg = get_arg_in_braces(ses, arg, left, GET_ALL);
+		      get_arg_in_braces(ses, arg, temp, GET_ALL);
 
 		if (*temp)
 		{
@@ -97,7 +171,7 @@ DO_COMMAND(do_echo)
 
 			substitute(ses, left, temp, SUB_COL|SUB_ESC);
 
-			do_one_prompt(ses, temp, row);
+			split_show(ses, temp, row, 0);
 
 			return ses;
 		}
@@ -123,15 +197,17 @@ DO_COMMAND(do_echo)
 		if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
 		{
 			save_pos(ses);
-			goto_rowcol(ses, ses->bot_row, 1);
-		}
 
-		printline(ses, &output, lnf);
+			goto_pos(ses, ses->split->bot_row, 1);
 
-		if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
-		{
+			print_line(ses, &output, lnf);
+			
 			restore_pos(ses);
 		}
+		else
+		{
+			print_line(ses, &output, lnf);
+		}
 	}
 	str_free(output);
 
@@ -141,13 +217,11 @@ DO_COMMAND(do_echo)
 
 DO_COMMAND(do_end)
 {
-	char left[BUFFER_SIZE];
-
 	if (*arg)
 	{
-		sub_arg_in_braces(ses, arg, left, GET_ALL, SUB_VAR|SUB_FUN|SUB_COL|SUB_ESC|SUB_LNF);
+		sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN|SUB_COL|SUB_ESC|SUB_LNF);
 
-		quitmsg(left);
+		quitmsg(arg1);
 	}
 	else
 	{
@@ -157,44 +231,6 @@ DO_COMMAND(do_end)
 }
 
 
-DO_COMMAND(do_forall)
-{
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], temp[BUFFER_SIZE];
-
-	arg = sub_arg_in_braces(ses, arg, left, GET_ALL, SUB_VAR|SUB_FUN);
-	arg = get_arg_in_braces(ses, arg, right, TRUE);
-
-	tintin_printf2(ses, "\e[1;31m#NOTICE: The #forall command is going to be removed in the next release, please switch to the #foreach command.\n");
-
-	if (*left == 0 || *right == 0)
-	{
-		tintin_printf2(ses, "#ERROR: #FORALL - PROVIDE 2 ARGUMENTS");
-	}
-	else
-	{
-		arg = left;
-
-		while (*arg)
-		{
-			arg = get_arg_in_braces(ses, arg, temp, TRUE);
-
-			RESTRING(gtd->cmds[0], temp);
-
-			substitute(ses, right, temp, SUB_CMD);
-
-			ses = script_driver(ses, LIST_COMMAND, temp);
-
-			if (*arg == COMMAND_SEPARATOR)
-			{
-				arg++;
-			}
-
-		}
-	}
-	return ses;
-}
-
-
 DO_COMMAND(do_nop)
 {
 	return ses;
@@ -203,126 +239,57 @@ DO_COMMAND(do_nop)
 
 DO_COMMAND(do_send)
 {
-	char left[BUFFER_SIZE];
-
 	push_call("do_send(%p,%p)",ses,arg);
 
-	get_arg_in_braces(ses, arg, left, TRUE);
+	get_arg_in_braces(ses, arg, arg1, GET_ALL);
 
-	write_mud(ses, left, SUB_VAR|SUB_FUN|SUB_ESC|SUB_EOL);
+	write_mud(ses, arg1, SUB_VAR|SUB_FUN|SUB_ESC|SUB_EOL);
 
 	pop_call();
 	return ses;
 }
 
 
-DO_COMMAND(do_showme)
-{
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], temp[STRING_SIZE], *output;
-	int lnf;
-
-	arg = get_arg_in_braces(ses, arg, left, TRUE);
-
-	lnf = !str_suffix(left, "\\");
-
-	substitute(ses, left, temp, SUB_VAR|SUB_FUN);
-	substitute(ses, temp, left, SUB_COL|SUB_ESC);
-
-	arg = sub_arg_in_braces(ses, arg, right, GET_ONE, SUB_VAR|SUB_FUN);
-
-	do_one_line(left, ses);
-
-	if (HAS_BIT(ses->flags, SES_FLAG_GAG))
-	{
-		DEL_BIT(ses->flags, SES_FLAG_GAG);
-
-		return ses;
-	}
-
-	if (*right)
-	{
-		do_one_prompt(ses, left, (int) get_number(ses, right));
-
-		return ses;
-	}
-
-	if (strip_vt102_strlen(ses, ses->more_output) != 0)
-	{
-		output = str_dup_printf("\n\e[0m%s\e[0m", left);
-	}
-	else
-	{
-		output = str_dup_printf("\e[0m%s\e[0m", left);
-	}
-
-	add_line_buffer(ses, output, lnf);
-
-	if (ses == gtd->ses)
-	{
-		if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
-		{
-			save_pos(ses);
-			goto_rowcol(ses, ses->bot_row, 1);
-		}
-
-		printline(ses, &output, lnf);
-
-		if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
-		{
-			restore_pos(ses);
-		}
-	}
-
-	str_free(output);
-
-	return ses;
-}
-
 
 DO_COMMAND(do_test)
 {
-	long long x, time1, time2;
-	char *str, buf[100001];
-
-	tintin_printf2(ses, "  input_level %d", gtd->input_level);
-	tintin_printf2(ses, "verbose_level %d", gtd->verbose_level);
-	tintin_printf2(ses, "  quiet_level %d", gtd->quiet);
-	tintin_printf2(ses, "  debug_level %d", gtd->debug_level);
-
-	return ses;
-
-	tintin_printf2(ses, "testing");
-
-	buf[0] = 0;
-
-	time1 = utime();
-
-	str = buf;
-
-	for (x = 0 ; x < 100000 ; x++)
+	if (!strcmp(arg, "string"))
 	{
-		*str++ = '*';
-	}
-	*str = 0;
+		char *test = str_dup("\e[32m0 2 4 6 8 A C E");
 
-	time2 = utime();
+		test = str_ins_str(ses, &test, "\e[33mbli bli", 4, 11);
 
-	tintin_printf(ses, "time for strcat: %lld %d", time2 - time1, strlen(buf));
+		printf("test: [%s]\n", test);
 
-	str = str_dup("");
-
-	time1 = utime();
-
-	for (x = 0 ; x < 100000 ; x++)
-	{
-		str_cpy_printf(&str, "%c", '*');
+		str_free(test);
+		
+		return ses;
 	}
 
-	time2 = utime();
+	strcpy(arg2, "9");
+	strcpy(arg3, "<f0b8>");
+	strcpy(arg4, "1 9");
 
-	tintin_printf(ses, "time for str_cat: %lld %d", time2 - time1, strlen(str));
+	if (isdigit((int) arg[0]))
+	{
+		sprintf(arg2, "%d", (arg[0] - '0') * (arg[0] - '0'));
 
-	str_free(str);
+		if ((isxdigit((int) arg[1]) && isxdigit((int) arg[2]) && isxdigit((int) arg[3])) || (arg[1] == '?' && arg[2] == '?' && arg[3] == '?'))
+		{
+			sprintf(arg3, "<f%c%c%c>", arg[1], arg[2], arg[3]);
+
+			if (isdigit((int) arg[4]) && isdigit((int) arg[5]))
+			{
+				sprintf(arg4, "%f %d %s", (arg[4] - '0') * (arg[4] - '0') / 10.0, (arg[5] - '0') * (arg[5] - '0'), &arg[6]);
+
+				tintin_printf2(ses, "do_test debug: %s", arg4);
+			}
+		}
+	}
+	sprintf(arg1, "#line quiet {#event {RECEIVED KEYPRESS} {#end \\};#screen cursor hide;#screen clear all;#event {SECOND} #loop 0 %s cnt #delay {$cnt / (1.0+%s)} #draw %s rain 1 1 -1 -1 rain %s}", arg2, arg2, arg3, arg4);
+
+	script_driver(gtd->ses, LIST_COMMAND, arg1);
 
 	return ses;
 }
+
