@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,10 +13,10 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
-*******************************************************************************/
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
+******************************************************************************/
 
 /******************************************************************************
 *                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
@@ -28,17 +27,15 @@
 
 #include "tintin.h"
 
-#include <errno.h>
-#include <sys/types.h>
+#include <arpa/inet.h>
+#include <fcntl.h>
 #include <netinet/in.h>
 #include <netdb.h>
 #include <signal.h>
-#include <arpa/inet.h>
-#include <sys/socket.h>
-#include <fcntl.h>
+#include <sys/types.h>
 
 /*
-	IPv6 compatible connect code, doesn't work on several platforms.
+	IPv6 compatible connect code.
 */
 
 #ifdef HAVE_GETADDRINFO
@@ -71,7 +68,7 @@ int connect_mud(struct session *ses, char *host, char *port)
 
 		if (error)
 		{
-			tintin_printf(ses, "#SESSION '%s' COULD NOT CONNECT - UNKNOWN HOST.", ses->name);
+			tintin_printf2(ses, "#SESSION '%s' COULD NOT CONNECT - UNKNOWN HOST.", ses->name);
 
 			return -1;
 		}
@@ -81,27 +78,49 @@ int connect_mud(struct session *ses, char *host, char *port)
 
 	if (sock < 0)
 	{
-		syserr("socket");
-	}
+		syserr_printf(ses, "connect_mud: socket");
 
-	ses->connect_error = connect(sock, address->ai_addr, address->ai_addrlen);
-
-	if (ses->connect_error)
-	{
-		close(sock);
 		freeaddrinfo(address);
 
-		return 0;
+		return -1;
 	}
 
+//	fcntl(sock, F_SETFL, O_NDELAY);
+
+        ses->connect_error = connect(sock, address->ai_addr, address->ai_addrlen);
+
+        if (ses->connect_error)
+        {
+                syserr_printf(ses, "connect_mud: connect");
+
+                close(sock);
+
+                freeaddrinfo(address);
+
+                return -1;
+        }
+
 	if (fcntl(sock, F_SETFL, O_NDELAY|O_NONBLOCK) == -1)
 	{
-		perror("connect_mud: fcntl O_NDELAY|O_NONBLOCK");
+		syserr_printf(ses, "connect_mud: fcntl O_NDELAY|O_NONBLOCK");
+
+		close(sock);
+
+		freeaddrinfo(address);
+
+		return -1;
 	}
 
-	getnameinfo(address->ai_addr, address->ai_addrlen, ip, 100, NULL, 0, NI_NUMERICHOST);
+	error = getnameinfo(address->ai_addr, address->ai_addrlen, ip, 100, NULL, 0, NI_NUMERICHOST);
 
-	RESTRING(ses->session_ip, ip);
+	if (error)
+	{
+		syserr_printf(ses, "connect_mud: getnameinfo:");
+	}
+	else
+	{
+		RESTRING(ses->session_ip, ip);
+	}
 
 	freeaddrinfo(address);
 
@@ -115,6 +134,8 @@ int connect_mud(struct session *ses, char *host, char *port)
 	int sock, d;
 	struct sockaddr_in sockaddr;
 
+	print_stdout("debug: NO ADDRESS INFO?\n");
+
 	if (sscanf(host, "%d.%d.%d.%d", &d, &d, &d, &d) == 4)
 	{
 		sockaddr.sin_addr.s_addr = inet_addr(host);
@@ -144,7 +165,9 @@ int connect_mud(struct session *ses, char *host, char *port)
 
 	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
 	{
-		syserr("socket");
+		syserr_printf(ses, "old_connect_mud: socket()");
+
+		return -1;
 	}
 
 	sockaddr.sin_family = AF_INET;
@@ -153,6 +176,8 @@ int connect_mud(struct session *ses, char *host, char *port)
 
 	if (ses->connect_error)
 	{
+		syserr_printf(ses, "connect_mud: connect");
+
 		close(sock);
 
 		return 0;
@@ -160,10 +185,10 @@ int connect_mud(struct session *ses, char *host, char *port)
 
 	if (fcntl(sock, F_SETFL, O_NDELAY|O_NONBLOCK) == -1)
 	{
-		perror("connect_mud: fcntl O_NDELAY|O_NONBLOCK");
+		syserr_printf(ses, "connect_mud: fcntl O_NDELAY|O_NONBLOCK");
 	}
 
-	RESTRING(ses->ip, inet_ntoa(sockaddr.sin_addr));
+	RESTRING(ses->session_ip, inet_ntoa(sockaddr.sin_addr));
 
 	return sock;
 }
@@ -172,14 +197,28 @@ int connect_mud(struct session *ses, char *host, char *port)
 
 void write_line_mud(struct session *ses, char *line, int size)
 {
-	static int retry;
+	int result;
 
 	push_call("write_line_mud(%p,%p)",line,ses);
 
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "SEND OUTPUT", line, ntos(size));
+
+	if (check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "CATCH SEND OUTPUT", line, ntos(size)))
+	{
+		pop_call();
+		return;
+	}
+
 	if (ses == gts)
 	{
-		tintin_printf2(ses, "#NO SESSION ACTIVE. USE: %csession {name} {host} {port} TO START ONE.", gtd->tintin_char);
-
+		if (HAS_BIT(gtd->flags, TINTIN_FLAG_CHILDLOCK))
+		{
+			tintin_printf2(ses, "#THIS SESSION IS CHILD LOCKED, PRESS CTRL-D TO EXIT.");
+		}
+		else
+		{
+			tintin_printf2(ses, "#NO SESSION ACTIVE. USE: %csession {name} {host} {port} TO START ONE.", gtd->tintin_char);
+		}
 		pop_call();
 		return;
 	}
@@ -192,45 +231,54 @@ void write_line_mud(struct session *ses, char *line, int size)
 		return;
 	}
 
-#ifdef HAVE_GNUTLS_H
 
-	if (ses->ssl)
+	if (!HAS_BIT(ses->telopts, TELOPT_FLAG_TELNET) && HAS_BIT(ses->charset, CHARSET_FLAG_ALL_TOUTF8))
 	{
-		int result;
+		char buf[BUFFER_SIZE];
 
-		result = gnutls_record_send(ses->ssl, line, size);
+		size = utf8_to_all(ses, line, buf);
 
-		while (result == GNUTLS_E_INTERRUPTED || result == GNUTLS_E_AGAIN)
-		{
-			result = gnutls_record_send(ses->ssl, 0, 0);
-		}
+		memcpy(line, buf, size);
+
+		line[size] = 0;
 	}
-	else
 
-#endif
-
-	if (write(ses->socket, line, size) == -1)
 	{
-		if (retry++ < 10)
+		if (ses->mccp3)
 		{
-			usleep(100000);
+			result = client_write_compressed(ses, line, size);
+		}
+#ifdef HAVE_GNUTLS_H
+		else if (ses->ssl)
+		{
+			result = gnutls_record_send(ses->ssl, line, size);
 
-			write_line_mud(ses, line, size);
+			while (result == GNUTLS_E_INTERRUPTED || result == GNUTLS_E_AGAIN)
+			{
+				result = gnutls_record_send(ses->ssl, 0, 0);
+			}
+		}
+#endif
+		else
+		{
+			result = write(ses->socket, line, size);
+
+			if (result == -1)
+			{
+				syserr_printf(ses, "write_line_mud: write");
+			}
+		}
+		
+		if (result == -1)
+		{
+			cleanup_session(ses);
 
 			pop_call();
 			return;
 		}
-		perror("write in write_line_mud");
-
-		cleanup_session(ses);
-
-		pop_call();
-		return;
 	}
 
-	retry = 0;
-
-	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "SEND OUTPUT", line);
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "SENT OUTPUT", line, ntos(size));
 
 	pop_call();
 	return;
@@ -268,36 +316,75 @@ int read_buffer_mud(struct session *ses)
 		pop_call();
 		return FALSE;
 	}
-	ses->read_len = translate_telopts(ses, buffer, size);
+	ses->read_len = client_translate_telopts(ses, buffer, size);
 
 	pop_call();
 	return TRUE;
 }
 
+int detect_prompt(struct session *ses, char *original)
+{
+	char strip[BUFFER_SIZE];
+	struct listroot *root = ses->list[LIST_PROMPT];
+	struct listnode *node;
+
+	if (HAS_BIT(ses->charset, CHARSET_FLAG_ALL_TOUTF8))
+	{
+		all_to_utf8(ses, original, strip);
+
+		strcpy(original, strip);
+	}
+
+	strip_vt102_codes(original, strip);
+
+	for (root->update = 0 ; root->update < root->used ; root->update++)
+	{
+		node = root->list[root->update];
+
+		if (check_one_regexp(ses, node, strip, original, 0))
+		{
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
 
 void readmud(struct session *ses)
 {
 	char *line, *next_line;
-	char linebuf[STRING_SIZE];
+	char linebuf[BUFFER_SIZE];
+	int len;
+	struct session *cts;
 
 	push_call("readmud(%p)", ses);
 
+	gtd->mud_output_len = 0;
+
 	if (gtd->mud_output_len < BUFFER_SIZE)
 	{
 		check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "RECEIVED OUTPUT", gtd->mud_output_buf);
+
+		if (check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "CATCH RECEIVED OUTPUT", gtd->mud_output_buf))
+		{
+			pop_call();
+			return;
+		}
 	}
 
-	gtd->mud_output_len = 0;
-
 	/* separate into lines and print away */
 
+	// cts = current tintin session, may have to make this global to avoid glitches
+
+	cts = gtd->ses;
+
 	if (HAS_BIT(gtd->ses->flags, SES_FLAG_SPLIT))
 	{
 		save_pos(gtd->ses);
-		goto_rowcol(gtd->ses, gtd->ses->bot_row, 1);
+
+		goto_pos(gtd->ses, gtd->ses->split->bot_row, 1);
 	}
 
-	SET_BIT(gtd->ses->flags, SES_FLAG_READMUD);
+	SET_BIT(cts->flags, SES_FLAG_READMUD);
 
 	for (line = gtd->mud_output_buf ; line && *line ; line = next_line)
 	{
@@ -305,23 +392,61 @@ void readmud(struct session *ses)
 
 		if (next_line)
 		{
-			*next_line = 0;
-			next_line++;
-		}
-		else if (*line == 0)
-		{
-			break;
-		}
-
-		if (next_line == NULL && strlen(ses->more_output) < BUFFER_SIZE / 2)
-		{
-			if (!HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_PROMPT))
+			if (next_line - line >= BUFFER_SIZE / 3)
 			{
-				if (gts->check_output)
+				// This is not ideal, but being a rare case it'll suffice for now
+
+				next_line = &line[BUFFER_SIZE / 3];
+			}
+			*next_line++ = 0;
+		}
+		else
+		{
+			if (*line == 0)
+			{
+				break;
+			}
+
+			len = strlen(line);
+
+			if (strlen(line) > BUFFER_SIZE / 3)
+			{
+				len = BUFFER_SIZE / 3;
+
+				next_line = &line[len];
+
+				*next_line++ = 0;
+			}
+
+			if (strlen(ses->more_output) < BUFFER_SIZE / 3)
+			{
+				if (!HAS_BIT(ses->telopts, TELOPT_FLAG_PROMPT))
 				{
-					strcat(ses->more_output, line);
-					ses->check_output = utime() + gts->check_output;
-					break;
+					if (ses->packet_patch)
+					{
+						strcat(ses->more_output, line);
+						ses->check_output = utime() + ses->packet_patch;
+
+						break;
+					}
+					else if (HAS_BIT(ses->flags, SES_FLAG_AUTOPATCH))
+					{
+						if (ses->list[LIST_PROMPT]->list[0])
+						{
+							if (!detect_prompt(ses, line))
+							{
+								strcat(ses->more_output, line);
+								ses->check_output = utime() + 500000ULL;
+								break;
+							}
+						}
+						else if (HAS_BIT(ses->flags, SES_FLAG_AUTOPROMPT))
+						{
+							strcat(ses->more_output, line);
+							ses->check_output = utime() + 500000ULL;
+							break;
+						}
+					}
 				}
 			}
 		}
@@ -330,7 +455,8 @@ void readmud(struct session *ses)
 		{
 			if (ses->check_output)
 			{
-				sprintf(linebuf, "%s%s", ses->more_output, line);
+				strcat(ses->more_output, line);
+				strcpy(linebuf, ses->more_output);
 
 				ses->more_output[0] = 0;
 			}
@@ -344,9 +470,20 @@ void readmud(struct session *ses)
 			strcpy(linebuf, line);
 		}
 
-		process_mud_output(ses, linebuf, next_line == NULL);
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_ALL_TOUTF8))
+		{
+			char tempbuf[BUFFER_SIZE];
+
+			all_to_utf8(ses, linebuf, tempbuf);
+
+			process_mud_output(ses, tempbuf, next_line == NULL);
+		}
+		else
+		{
+			process_mud_output(ses, linebuf, next_line == NULL);
+		}
 	}
-	DEL_BIT(gtd->ses->flags, SES_FLAG_READMUD);
+	DEL_BIT(cts->flags, SES_FLAG_READMUD);
 
 	if (HAS_BIT(gtd->ses->flags, SES_FLAG_SPLIT))
 	{
@@ -361,25 +498,39 @@ void readmud(struct session *ses)
 void process_mud_output(struct session *ses, char *linebuf, int prompt)
 {
 	char line[STRING_SIZE];
+	int str_len, raw_len;
 
 	push_call("process_mud_output(%p,%p,%d)",ses,linebuf,prompt);
 
 	ses->check_output = 0;
 
-	strip_vt102_codes(linebuf, line);
+	raw_len = strlen(linebuf);
+	str_len = strip_vt102_codes(linebuf, line);
 
 	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "RECEIVED LINE", linebuf, line);
 
-	if (prompt)
+	if (check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "CATCH RECEIVED LINE", linebuf, line))
 	{
-		check_all_events(ses, SUB_ARG|SUB_SEC, 0, 2, "RECEIVED PROMPT", linebuf, line);
+		pop_call();
+		return;
+	}
+
+	if (str_len && prompt)
+	{
+		check_all_events(ses, SUB_ARG|SUB_SEC, 0, 4, "RECEIVED PROMPT", linebuf, line, ntos(raw_len), ntos(str_len));
+
+		if (check_all_events(ses, SUB_ARG|SUB_SEC, 0, 4, "CATCH RECEIVED PROMPT", linebuf, line, ntos(raw_len), ntos(str_len)))
+		{
+			pop_call();
+			return;
+		}
 	}
 
 	if (HAS_BIT(ses->flags, SES_FLAG_COLORPATCH))
 	{
-		sprintf(line, "%s%s%s", ses->color, linebuf, "\e[0m");
+		sprintf(line, "%s%s%s", ses->color_patch, linebuf, "\e[0m");
 
-		get_color_codes(ses->color, linebuf, ses->color);
+		get_color_codes(ses->color_patch, linebuf, ses->color_patch, GET_ALL);
 
 		linebuf = line;
 	}
@@ -390,15 +541,17 @@ void process_mud_output(struct session *ses, char *linebuf, int prompt)
 		Take care of gags, vt102 support still goes
 	*/
 
-	if (HAS_BIT(ses->flags, SES_FLAG_GAG))
+	if (ses->gagline > 0)
 	{
-		strip_non_vt102_codes(linebuf, ses->more_output);
+		ses->gagline--;
 
-		printf("%s", ses->more_output);
+		strip_non_vt102_codes(linebuf, line);
 
-		ses->more_output[0] = 0;
+		print_stdout("%s", line);
 
-		DEL_BIT(ses->flags, SES_FLAG_GAG);
+		strip_vt102_codes(linebuf, line);
+
+		show_debug(ses, LIST_GAG, "#DEBUG GAG {%d} {%s}", ses->gagline + 1, line);
 
 		pop_call();
 		return;
@@ -410,17 +563,18 @@ void process_mud_output(struct session *ses, char *linebuf, int prompt)
 	{
 		char *output = str_dup(linebuf);
 
-		printline(ses, &output, prompt);
+		print_line(ses, &output, prompt);
 
 		str_free(output);
-	}
-	else if (HAS_BIT(ses->flags, SES_FLAG_SNOOP))
-	{
-		strip_vt102_codes_non_graph(linebuf, linebuf);
 
-		tintin_printf2(gtd->ses, "[%s] %s", ses->name, linebuf);
+		if (prompt)
+		{
+			if (!IS_SPLIT(ses))
+			{
+				ses->input->off = 1 + strip_vt102_strlen(ses, linebuf);
+			}
+		}
 	}
 	pop_call();
 	return;
 }
-
