@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,19 +13,146 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
-*******************************************************************************/
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
+******************************************************************************/
 
 /******************************************************************************
 *                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
 *                                                                             *
 *                         coded by Peter Unold 1992                           *
+*                     recoded by Igor van den Hoven 2005                      *
 ******************************************************************************/
 
 #include "tintin.h"
 
+// whether str1 is an abbreviation of str2
+
+int case_table[256] =
+{
+	  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,  15,
+	 16,  17,  18,  19,  20,  21,  22,  23,  24,  25,  26,  27,  28,  29,  30,  31,
+	 95,  33,  34,  35,  36,  37,  38,  39,  40,  41,  42,  43,  44,  45,  46,  47,
+	 48,  49,  50,  51,  52,  53,  54,  55,  56,  57,  58,  59,  60,  61,  62,  63,
+	 64,
+	 65,  66,  67,  68,  69,  70,  71,  72,  73,  74,  75,  76,  77,  78,  79,  80,  81,  82,  83,  84,  85,  86,  87,  88,  89,  90,
+	 91,  92,  93,  94,
+	 95,  96,
+	 65,  66,  67,  68,  69,  70,  71,  72,  73,  74,  75,  76,  77,  78,  79,  80,  81,  82,  83,  84,  85,  86,  87,  88,  89,  90,
+	123, 124, 125, 126, 127,
+	128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143,
+	144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159,
+	160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175,
+	176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191,
+	192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207,
+	208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223,
+	224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239,
+	240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255
+};
+
+int is_abbrev(char *str1, char *str2)
+{
+	char *str3 = gtd->is_result;
+
+	if (*str1 == 0)
+	{
+		return false;
+	}
+
+	if (*str2 == 0)
+	{
+		tintin_printf2(gtd->ses, "\e[1;31mis_abbrev(%s,%s)", str1, str2);
+
+		dump_stack();
+
+		return false;
+	}
+
+	while (true)
+	{
+		if (*str1 == 0)
+		{
+			strcpy(str3, str2);
+
+			return true;
+		}
+
+		if (case_table[(int) *str1] != case_table[(int) *str2])
+		{
+			return false;
+		}
+		str1++;
+		*str3++ = *str2++;
+	}
+}
+
+void filename_string(char *input, char *output)
+{
+	while (*input)
+	{
+		*output++ = (char) case_table[(int) *input++];
+	}
+	*output = 0;
+}
+
+int is_suffix(char *str1, char *str2)
+{
+	int len1, len2;
+
+	len1 = strlen(str1);
+	len2 = strlen(str2);
+
+	if (len1 >= len2)
+	{
+		if (is_abbrev(str1 + len1 - len2, str2))
+		{
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
+int is_vowel(char *str)
+{
+	switch (case_table[(int) *str])
+	{
+		case 'a':
+		case 'e':
+		case 'i':
+		case 'o':
+		case 'u':
+			return TRUE;
+	}
+	return FALSE;
+}
+
+struct session *execute(struct session *ses, char *format, ...)
+{
+	char *buffer;
+	va_list args;
+
+	va_start(args, format);
+	vasprintf(&buffer, format, args);
+	va_end(args);
+
+	if (*buffer)
+	{
+		if (*buffer != gtd->tintin_char)
+		{
+			*buffer = gtd->tintin_char;
+		}
+		get_arg_all(ses, buffer, buffer, FALSE);
+	}
+
+	ses = script_driver(ses, LIST_COMMAND, buffer);
+
+	free(buffer);
+
+	return ses;
+}
+
+
 struct session *parse_input(struct session *ses, char *input)
 {
 	char *line;
@@ -45,12 +171,16 @@ struct session *parse_input(struct session *ses, char *input)
 	{
 		line = (char *) malloc(BUFFER_SIZE);
 
-		strcpy(line, input);
+		sub_arg_all(ses, input, line, 1, SUB_SEC);
 
 		if (check_all_aliases(ses, line))
 		{
 			ses = script_driver(ses, LIST_ALIAS, line);
 		}
+		else if (HAS_BIT(ses->flags, SES_FLAG_SPEEDWALK) && is_speedwalk(ses, line))
+		{
+			process_speedwalk(ses, line);
+		}
 		else
 		{
 			write_mud(ses, line, SUB_EOL);
@@ -76,7 +206,7 @@ struct session *parse_input(struct session *ses, char *input)
 	{
 		input = space_out(input);
 
-		input = get_arg_all(ses, input, line, FALSE);
+		input = get_arg_all(ses, input, line, 0);
 
 		if (parse_command(ses, line))
 		{
@@ -118,7 +248,7 @@ struct session *parse_command(struct session *ses, char *input)
 
 	push_call("parse_command(%p,%p)",ses,input);
 
-	arg = get_arg_stop_spaces(ses, input, cmd1, 0);
+	arg = get_arg_stop_spaces(ses, input, cmd1, GET_ONE);
 
 	substitute(ses, cmd1, cmd2, SUB_VAR|SUB_FUN);
 
@@ -136,9 +266,74 @@ struct session *parse_command(struct session *ses, char *input)
 	return ses;
 }
 
+char *substitute_speedwalk(struct session *ses, char *input, char *output)
+{
+	char num[NUMBER_SIZE], name[BUFFER_SIZE], *pti, *ptn, *pto;
+	int cnt, max;
+
+	pti = input;
+	pto = output;
+
+	while (*pti && pto - output < INPUT_SIZE)
+	{
+		while (isspace((int) *pti))
+		{
+			pti++;
+		}
+
+		if (isdigit((int) *pti))
+		{
+			ptn = num;
+
+			while (isdigit((int) *pti))
+			{
+				if (ptn - num < 4)
+				{
+					*ptn++ = *pti++;
+				}
+				else
+				{
+					pti++;
+				}
+			}
+			*ptn = 0;
+
+			max = atoi(num);
+		}
+		else
+		{
+			max = 1;
+		}
+
+		pti = get_arg_stop_digits(ses, pti, name, GET_ONE);
+
+		if (*name == 0)
+		{
+			break;
+		}
+
+		for (cnt = 0 ; cnt < max ; cnt++)
+		{
+			if (output != pto)
+			{
+				*pto++ = COMMAND_SEPARATOR;
+			}
+			pto += sprintf(pto, "%s", name);
+		}
+
+		if (*pti == COMMAND_SEPARATOR)
+		{
+			pti++;
+		}
+	}
+	*pto = 0;
+
+	return output;
+}
+
 int is_speedwalk(struct session *ses, char *input)
 {
-	int flag = FALSE;
+	int digit = 0, flag = FALSE;
 
 	while (*input)
 	{
@@ -150,7 +345,12 @@ int is_speedwalk(struct session *ses, char *input)
 			case 'w':
 			case 'u':
 			case 'd':
-				flag = TRUE;
+				if (digit > 3)
+				{
+					return FALSE;
+				}
+				digit = 0;
+				flag  = TRUE;
 				break;
 
 			case '0':
@@ -163,6 +363,7 @@ int is_speedwalk(struct session *ses, char *input)
 			case '7':
 			case '8':
 			case '9':
+				digit++;
 				flag = FALSE;
 				break;
 
@@ -215,7 +416,7 @@ struct session *parse_tintin_command(struct session *ses, char *input)
 	char line[BUFFER_SIZE];
 	struct session *sesptr;
 
-	input = get_arg_stop_spaces(ses, input, line, 0);
+	input = get_arg_stop_spaces(ses, input, line, GET_ONE);
 
 	substitute(ses, line, line, SUB_VAR|SUB_FUN);
 
@@ -223,7 +424,7 @@ struct session *parse_tintin_command(struct session *ses, char *input)
 	{
 		int cnt = atoi(line);
 
-		input = get_arg_in_braces(ses, input, line, TRUE);
+		input = get_arg_in_braces(ses, input, line, GET_ALL);
 
 		while (cnt-- > 0)
 		{
@@ -238,7 +439,7 @@ struct session *parse_tintin_command(struct session *ses, char *input)
 	{
 		if (*input)
 		{
-			input = get_arg_in_braces(ses, input, line, TRUE);
+			input = get_arg_in_braces(ses, input, line, GET_ALL);
 
 			substitute(ses, line, line, SUB_VAR|SUB_FUN);
 
@@ -252,21 +453,51 @@ struct session *parse_tintin_command(struct session *ses, char *input)
 		}
 	}
 
-	tintin_printf(ses, "#ERROR: #UNKNOWN TINTIN-COMMAND '%s'.", line);
+	if (*line == '!')
+	{
+		show_error(ses, LIST_COMMAND, "#!%s %s", line + 1, input);
+
+		return ses;
+	}
+
+	tintin_printf2(ses, "#ERROR: #UNKNOWN TINTIN-COMMAND '%s'.", line);
+
+	check_all_events(ses, SUB_ARG|SUB_SEC, 0, 1, "UNKNOWN COMMAND", line);
 
 	return ses;
 }
 
 
+int cnt_arg_all(struct session *ses, char *string, int flag)
+{
+	char *arg, tmp[BUFFER_SIZE];
+	int cnt;
+
+	arg = string;
+	cnt = 0;
+
+	while (*arg)
+	{
+		cnt++;
+
+		arg = get_arg_in_braces(ses, arg, tmp, flag);
+
+		if (*arg == COMMAND_SEPARATOR)
+		{
+			arg++;
+		}
+	}
+	return cnt;
+}
+
 /*
 	get all arguments - only check for unescaped command separators
 */
 
-
 char *get_arg_all(struct session *ses, char *string, char *result, int verbatim)
 {
 	char *pto, *pti;
-	int nest = 0;
+	int skip, nest = 0;
 
 	pti = string;
 	pto = result;
@@ -284,13 +515,24 @@ char *get_arg_all(struct session *ses, char *string, char *result, int verbatim)
 
 	while (*pti)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
 			continue;
 		}
 
+		skip = find_secure_color_code(pti);
+
+		if (skip)
+		{
+			while (skip--)
+			{
+				*pto++ = *pti++;
+			}
+			continue;
+		}
+
 		if (*pti == '\\' && pti[1] == COMMAND_SEPARATOR)
 		{
 			*pto++ = *pti++;
@@ -308,12 +550,34 @@ char *get_arg_all(struct session *ses, char *string, char *result, int verbatim)
 			nest--;
 		}
 		*pto++ = *pti++;
+
+		if (pto - result >= BUFFER_SIZE - 3)
+		{
+			tintin_printf2(ses, "#ERROR: INPUT BUFFER OVERFLOW.");
+
+			pto--;
+
+			break;
+		}
 	}
 	*pto = '\0'; 
 
 	return pti;
 }
 
+char *sub_arg_all(struct session *ses, char *string, char *result, int verbatim, int sub)
+{
+	char *buffer = str_alloc(UMAX(strlen(string), BUFFER_SIZE));
+
+	string = get_arg_all(ses, string, buffer, verbatim);
+
+	substitute(ses, buffer, result, sub);
+
+	str_free(buffer);
+
+	return string;
+}
+
 /*
 	Braces are stripped in braced arguments leaving all else as is.
 */
@@ -321,7 +585,7 @@ char *get_arg_all(struct session *ses, char *string, char *result, int verbatim)
 char *get_arg_in_braces(struct session *ses, char *string, char *result, int flag)
 {
 	char *pti, *pto;
-	int nest = 1;
+	int skip, nest = 1;
 
 	pti = space_out(string);
 	pto = result;
@@ -343,14 +607,24 @@ char *get_arg_in_braces(struct session *ses, char *string, char *result, int fla
 
 	while (*pti)
 	{
-		
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
 			continue;
 		}
 
+		skip = find_secure_color_code(pti);
+
+		if (skip)
+		{
+			while (skip--)
+			{
+				*pto++ = *pti++;
+			}
+			continue;
+		}
+
 		if (*pti == DEFAULT_OPEN)
 		{
 			nest++;
@@ -369,7 +643,7 @@ char *get_arg_in_braces(struct session *ses, char *string, char *result, int fla
 
 	if (*pti == 0)
 	{
-		tintin_printf2(NULL, "#Unmatched braces error!");
+		tintin_printf2(ses, "#ERROR: GET BRACED ARGUMENT: UNMATCHED BRACE.");
 	}
 	else
 	{
@@ -382,7 +656,7 @@ char *get_arg_in_braces(struct session *ses, char *string, char *result, int fla
 
 char *sub_arg_in_braces(struct session *ses, char *string, char *result, int flag, int sub)
 {
-	char *buffer = str_alloc(UMAX(strlen(string), BUFFER_SIZE));
+	char *buffer = str_alloc(strlen(string) + BUFFER_SIZE);
 
 	string = get_arg_in_braces(ses, string, buffer, flag);
 
@@ -400,20 +674,31 @@ char *sub_arg_in_braces(struct session *ses, char *string, char *result, int fla
 char *get_arg_with_spaces(struct session *ses, char *string, char *result, int flag)
 {
 	char *pto, *pti;
-	int nest = 0;
+	int skip, nest = 0;
 
 	pti = space_out(string);
 	pto = result;
 
 	while (*pti)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
 			continue;
 		}
 
+		skip = find_secure_color_code(pti);
+
+		if (skip)
+		{
+			while (skip--)
+			{
+				*pto++ = *pti++;
+			}
+			continue;
+		}
+
 		if (*pti == '\\' && pti[1] == COMMAND_SEPARATOR)
 		{
 			*pto++ = *pti++;
@@ -442,6 +727,73 @@ char *get_arg_with_spaces(struct session *ses, char *string, char *result, int f
 */
 
 char *get_arg_stop_spaces(struct session *ses, char *string, char *result, int flag)
+{
+	char *pto, *pti;
+	int skip, nest = 0;
+
+	pti = space_out(string);
+	pto = result;
+
+	while (*pti)
+	{
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
+		{
+			*pto++ = *pti++;
+			*pto++ = *pti++;
+			continue;
+		}
+
+		skip = find_secure_color_code(pti);
+
+		if (skip)
+		{
+			while (skip--)
+			{
+				*pto++ = *pti++;
+			}
+			continue;
+		}
+
+
+		if (*pti == '\\' && pti[1] == COMMAND_SEPARATOR)
+		{
+			*pto++ = *pti++;
+		}
+		else if (*pti == COMMAND_SEPARATOR && nest == 0)
+		{
+			break;
+		}
+		else if (isspace((int) *pti) && nest == 0)
+		{
+			pti++;
+			break;
+		}
+		else if (*pti == DEFAULT_OPEN)
+		{
+			nest++;
+		}
+		else if (*pti == '[' && HAS_BIT(flag, GET_NST))
+		{
+			nest++;
+		}
+		else if (*pti == DEFAULT_CLOSE)
+		{
+			nest--;
+		}
+		else if (*pti == ']' && HAS_BIT(flag, GET_NST))
+		{
+			nest--;
+		}
+		*pto++ = *pti++;
+	}
+	*pto = '\0';
+
+	return pti;
+}
+
+// Get one arg, stop at numbers, used for speedwalks
+
+char *get_arg_stop_digits(struct session *ses, char *string, char *result, int flag)
 {
 	char *pto, *pti;
 	int nest = 0;
@@ -451,7 +803,7 @@ char *get_arg_stop_spaces(struct session *ses, char *string, char *result, int f
 
 	while (*pti)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
@@ -466,9 +818,8 @@ char *get_arg_stop_spaces(struct session *ses, char *string, char *result, int f
 		{
 			break;
 		}
-		else if (isspace((int) *pti) && nest == 0)
+		else if (isdigit((int) *pti) && nest == 0)
 		{
-			pti++;
 			break;
 		}
 		else if (*pti == DEFAULT_OPEN)
@@ -508,7 +859,7 @@ char *space_out(char *string)
 }
 
 /*
-	For list handling
+	For variable handling
 */
 
 char *get_arg_to_brackets(struct session *ses, char *string, char *result)
@@ -522,7 +873,7 @@ char *get_arg_to_brackets(struct session *ses, char *string, char *result)
 
 	while (*pti)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
@@ -589,7 +940,7 @@ char *get_arg_at_brackets(struct session *ses, char *string, char *result)
 
 	while (*pti)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
@@ -620,7 +971,7 @@ char *get_arg_at_brackets(struct session *ses, char *string, char *result)
 
 	if (nest)
 	{
-		tintin_printf2(NULL, "#UNMATCHED AT BRACKETS ERROR.");
+		tintin_printf2(NULL, "#ERROR: GET BRACKETED VARIABLE: UNMATCHED BRACKET.");
 	}
 	*pto = 0;
 
@@ -646,7 +997,7 @@ char *get_arg_in_brackets(struct session *ses, char *string, char *result)
 
 	while (*pti)
 	{
-		if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && *pti & 128 && pti[1] != 0)
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
 		{
 			*pto++ = *pti++;
 			*pto++ = *pti++;
@@ -671,7 +1022,7 @@ char *get_arg_in_brackets(struct session *ses, char *string, char *result)
 
 	if (*pti == 0)
 	{
-		tintin_printf2(NULL, "#UNMATCHED IN BRACKETS ERROR.");
+		tintin_printf2(NULL, "#ERROR: GET BRACKETED ARGUMENT: UNMATCHED BRACKET.");
 	}
 	else
 	{
@@ -682,8 +1033,28 @@ char *get_arg_in_brackets(struct session *ses, char *string, char *result)
 	return pti;
 }
 
+char *get_char(struct session *ses, char *string, char *result)
+{
+	char *pti = string;
+
+	if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, pti))
+	{
+		pti += sprintf(result, "%c%c", pti[0], pti[1]);
+	}
+	else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(pti))
+	{
+		pti += sprintf(result, "%.*s", get_utf8_size(pti), pti);
+	}
+	else
+	{
+		pti += sprintf(result, "%c", pti[0]);
+	}
+
+	return pti;
+}
+
 /*
-	send command+argument to the mud
+	send command to the mud
 */
 
 void write_mud(struct session *ses, char *command, int flags)
@@ -693,22 +1064,19 @@ void write_mud(struct session *ses, char *command, int flags)
 
 	size = substitute(ses, command, output, flags);
 
-	if (HAS_BIT(ses->flags, SES_FLAG_MAPPING))
+	if (gtd->level->ignore == 0 && HAS_BIT(ses->flags, SES_FLAG_PATHMAPPING))
 	{
 		if (ses->map == NULL || ses->map->nofollow == 0)
 		{
-			check_insert_path(command, ses);
+			check_append_path(ses, command, NULL, 1);
 		}
 	}
 
-	if (ses->map && ses->map->in_room && ses->map->nofollow == 0)
+	if (gtd->level->ignore == 0 && ses->map && ses->map->in_room && ses->map->nofollow == 0)
 	{
-		if (!HAS_BIT(ses->map->flags, MAP_FLAG_NOFOLLOW))
+		if (follow_map(ses, command))
 		{
-			if (follow_map(ses, command))
-			{
-				return;
-			}
+			return;
 		}
 	}
 
@@ -716,23 +1084,24 @@ void write_mud(struct session *ses, char *command, int flags)
 }
 
 
-/******************************************************************************
-* do all of the functions to one line of buffer, VT102 codes and variables    *
-* substituted beforehand                                                      *
-******************************************************************************/
+/*
+	Check line for triggers
+*/
 
 void do_one_line(char *line, struct session *ses)
 {
-	char strip[BUFFER_SIZE];
+	char *strip;
 
-	push_call("[%s] do_one_line(%s)",ses->name,line);
+	push_call("do_one_line(%s,%p)",ses->name,line);
 
-	if (HAS_BIT(ses->flags, SES_FLAG_IGNORELINE))
+	if (gtd->level->ignore)
 	{
 		pop_call();
 		return;
 	}
 
+	strip = str_alloc_stack();
+
 	strip_vt102_codes(line, strip);
 
 	if (!HAS_BIT(ses->list[LIST_ACTION]->flags, LIST_FLAG_IGNORE))
@@ -742,7 +1111,7 @@ void do_one_line(char *line, struct session *ses)
 
 	if (!HAS_BIT(ses->list[LIST_PROMPT]->flags, LIST_FLAG_IGNORE))
 	{
-		check_all_prompts(ses, line, strip);
+		check_all_prompts(ses, line, strip, TRUE);
 	}
 
 	if (!HAS_BIT(ses->list[LIST_GAG]->flags, LIST_FLAG_IGNORE))
@@ -760,12 +1129,13 @@ void do_one_line(char *line, struct session *ses)
 		check_all_highlights(ses, line, strip);
 	}
 
-	if (HAS_BIT(ses->flags, SES_FLAG_LOGNEXT))
+	if (HAS_BIT(ses->logmode, LOG_FLAG_NEXT))
 	{
-		logit(ses, line, ses->lognext_file, TRUE);
+		logit(ses, line, ses->lognext_file, LOG_FLAG_LINEFEED);
 
-		DEL_BIT(ses->flags, SES_FLAG_LOGNEXT);
+		DEL_BIT(ses->logmode, LOG_FLAG_NEXT);
 	}
+
 	pop_call();
 	return;
 }
