@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,9 +13,9 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
 ******************************************************************************/
 
 /******************************************************************************
@@ -27,26 +26,22 @@
 
 #include "tintin.h"
 
-#include <errno.h>
-#include <signal.h>
-
-
 DO_COMMAND(do_all)
 {
-	char left[BUFFER_SIZE];
-	struct session *sesptr, *next_ses;
+	struct session *sesptr;
 
 	if (gts->next)
 	{
-		get_arg_in_braces(ses, arg, left, TRUE);
+		sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
 
-		substitute(ses, left, left, SUB_VAR|SUB_FUN);
-
-		for (sesptr = gts->next ; sesptr ; sesptr = next_ses)
+		for (sesptr = gts->next ; sesptr ; sesptr = gtd->all)
 		{
-			next_ses = sesptr->next;
+			gtd->all = sesptr->next;
 
-			script_driver(sesptr, LIST_COMMAND, left);
+			if (!HAS_BIT(sesptr->flags, SES_FLAG_CLOSED))
+			{
+				script_driver(sesptr, LIST_COMMAND, arg1);
+			}
 		}
 	}
 	else
@@ -59,7 +54,7 @@ DO_COMMAND(do_all)
 
 DO_COMMAND(do_session)
 {
-	char temp[BUFFER_SIZE], left[BUFFER_SIZE];
+	char temp[BUFFER_SIZE];
 	struct session *sesptr;
 	int cnt;
 
@@ -67,9 +62,9 @@ DO_COMMAND(do_session)
 
 	arg = temp;
 
-	arg = get_arg_in_braces(ses, arg, left,  FALSE);
+	arg = get_arg_in_braces(ses, arg, arg1, GET_ONE);
 
-	if (*left == 0)
+	if (*arg1 == 0)
 	{
 		tintin_puts(ses, "#THESE SESSIONS HAVE BEEN DEFINED:");
 
@@ -78,23 +73,51 @@ DO_COMMAND(do_session)
 			show_session(ses, sesptr);
 		}
 	}
-	else if (*left && *arg == 0)
+	else if (*arg1 && *arg == 0)
 	{
-		if (*left == '+')
+		if (!strncasecmp(arg1, "telnet://", 9))
+		{
+			char *pti, *pto;
+
+			pto = temp;
+			pti = arg1 + 9;
+
+			while (*pti)
+			{
+				if (*pti == '/')
+				{
+					break;
+				}
+				else if (*pti == ':')
+				{
+					pti++;
+					*pto++ = ' ';
+				}
+				else
+				{
+					*pto++ = *pti++;
+				}
+			}
+			*pto = 0;
+
+			ses = new_session(ses, "telnet", temp, 0, 0);
+		}
+
+		if (*arg1 == '+')
 		{
 			return activate_session(ses->next ? ses->next : gts->next ? gts->next : ses);
 		}
 
-		if (*left == '-')
+		if (*arg1 == '-')
 		{
 			return activate_session(ses->prev ? ses->prev : gts->prev ? gts->prev : ses);
 		}
 
-		if (is_number(left))
+		if (is_number(arg1))
 		{
 			for (cnt = 0, sesptr = gts ; sesptr ; cnt++, sesptr = sesptr->next)
 			{
-				if (cnt == atoi(left))
+				if (cnt == atoi(arg1))
 				{
 					return activate_session(sesptr);
 				}
@@ -105,26 +128,28 @@ DO_COMMAND(do_session)
 	}
 	else
 	{
-		ses = new_session(ses, left, arg, 0, 0);
+		ses = new_session(ses, arg1, arg, 0, 0);
 	}
-	return gtd->ses;
+	return ses;
 }
 
 
 DO_COMMAND(do_snoop)
 {
 	struct session *sesptr = ses;
-	char left[BUFFER_SIZE], right[BUFFER_SIZE];
 
-	sub_arg_in_braces(ses, arg, left, SUB_VAR|SUB_FUN, FALSE);
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg2, GET_ALL, SUB_VAR|SUB_FUN);
 
-	if (*left)
+	if (*arg1)
 	{
-		sesptr = find_session(left);
+		sesptr = find_session(arg1);
 
 		if (sesptr == NULL)
 		{
-			return show_error(ses, LIST_COMMAND, "#SNOOP: THERE'S NO SESSION NAMED {%s}", left);
+			show_error(ses, LIST_COMMAND, "#SNOOP: THERE'S NO SESSION NAMED {%s}", arg1);
+			
+			return ses;
 		}
 	}
 	else
@@ -132,7 +157,7 @@ DO_COMMAND(do_snoop)
 		sesptr = ses;
 	}
 
-	if (*right == 0)
+	if (*arg2 == 0)
 	{
 		if (HAS_BIT(sesptr->flags, SES_FLAG_SNOOP))
 		{
@@ -144,13 +169,13 @@ DO_COMMAND(do_snoop)
 		}
 		TOG_BIT(sesptr->flags, SES_FLAG_SNOOP);
 	}
-	else if (is_abbrev(right, "ON"))
+	else if (is_abbrev(arg2, "ON"))
 	{
 		show_message(ses, LIST_COMMAND, "#SNOOP: SNOOPING SESSION '%s'", sesptr->name);
 
 		SET_BIT(sesptr->flags, SES_FLAG_SNOOP);
 	}
-	else if (is_abbrev(right, "OFF"))
+	else if (is_abbrev(arg2, "OFF"))
 	{
 		show_message(ses, LIST_COMMAND, "#SNOOP: NO LONGER SNOOPING SESSION '%s'", sesptr->name);
 
@@ -167,15 +192,14 @@ DO_COMMAND(do_snoop)
 DO_COMMAND(do_zap)
 {
 	struct session *sesptr;
-	char left[BUFFER_SIZE];
 
 	push_call("do_zap(%p,%p)",ses,arg);
 
-	sub_arg_in_braces(ses, arg, left, GET_ALL, SUB_VAR|SUB_FUN);
+	sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
 
-	if (*left)
+	if (*arg1)
 	{
-		sesptr = find_session(left);
+		sesptr = find_session(arg1);
 
 		if (sesptr == NULL)
 		{
@@ -197,12 +221,21 @@ DO_COMMAND(do_zap)
 	if (sesptr == gts)
 	{
 		pop_call();
-		return do_end(NULL, "");
+		return execute(ses, "#END");
+//		do_end(NULL, "");
+	}
+
+	if (ses == sesptr)
+	{
+		cleanup_session(sesptr);
+
+		pop_call();
+		return gtd->ses;
 	}
 	cleanup_session(sesptr);
 
 	pop_call();
-	return gtd->ses;
+	return ses;
 }
 
 
@@ -212,15 +245,15 @@ void show_session(struct session *ses, struct session *ptr)
 
 	sprintf(temp, "%-10s %18s:%-5s", ptr->name, ptr->session_host, ptr->session_port);
 
-	strcat(temp, ptr == gtd->ses ? " (active)" :  "         ");
+	cat_sprintf(temp, " %8s", ptr == gtd->ses ? "(active)" :  "");
 
-	strcat(temp, ptr->mccp ? " (mccp)" : "       ");
+	cat_sprintf(temp, " %10s", ptr->mccp2 ? (ptr->mccp3 ? "(mccp 2+3)" : "(mccp 2)  ") : ptr->mccp3 ? "(mccp 3)" : "");
 
-	strcat(temp, HAS_BIT(ptr->flags, SES_FLAG_SNOOP) ? " (snoop)" : "        ");
+	cat_sprintf(temp, " %7s", HAS_BIT(ptr->flags, SES_FLAG_SNOOP) ? "(snoop)" : "");
 
-	strcat(temp, ptr->logfile ? " (log)" : "       ");
+	cat_sprintf(temp, " %5s", ptr->logfile ? "(log)" : "");
 
-	strcat(temp, ptr->ssl ? " (ssl)" : "      ");
+	cat_sprintf(temp, " %5s", ptr->ssl ? "(ssl)" : "");
 
 	tintin_puts2(ses, temp);
 }
@@ -236,6 +269,12 @@ struct session *find_session(char *name)
 			return ses;
 		}
 	}
+
+	if (!strcmp("ats", name))
+	{
+		return gtd->ses;
+	}
+
 	return NULL;
 }
 
@@ -265,7 +304,10 @@ struct session *activate_session(struct session *ses)
 
 	dirty_screen(ses);
 
-	tintin_printf(ses, "#SESSION '%s' ACTIVATED.", ses->name);
+	if (!check_all_events(ses, SUB_ARG, 0, 1, "GAG SESSION ACTIVATED", ses->name))
+	{
+		show_message(ses, LIST_COMMAND, "#SESSION '%s' ACTIVATED.", ses->name);
+	}
 
 	check_all_events(ses, SUB_ARG, 0, 1, "SESSION ACTIVATED", ses->name);
 
@@ -306,10 +348,7 @@ struct session *new_session(struct session *ses, char *name, char *arg, int desc
 
 		if (*port == 0)
 		{
-			tintin_puts(ses, "#HEY! SPECIFY A PORT NUMBER WILL YOU?");
-
-			pop_call();
-			return ses;
+			strcpy(port, "23");
 		}
 	}
 
@@ -327,12 +366,18 @@ struct session *new_session(struct session *ses, char *name, char *arg, int desc
 	newses->session_host  = strdup(host);
 	newses->session_ip    = strdup("");
 	newses->session_port  = strdup(port);
+	newses->created       = gtd->time;
 
 	newses->group         = strdup(gts->group);
 	newses->flags         = gts->flags;
+	newses->color         = gts->color;
+	newses->logmode       = gts->logmode;
+	newses->charset       = gts->charset;
+
 	newses->telopts       = gts->telopts;
 	newses->auto_tab      = gts->auto_tab;
-
+	newses->packet_patch  = gts->packet_patch;
+	newses->tab_width     = gts->tab_width;
 	newses->cmd_color     = strdup(gts->cmd_color);
 
 	newses->read_max      = gts->read_max;
@@ -340,6 +385,7 @@ struct session *new_session(struct session *ses, char *name, char *arg, int desc
 
 	newses->lognext_name  = strdup("");
 	newses->logline_name  = strdup("");
+	newses->rand          = utime();
 
 	LINK(newses, gts->next, gts->prev);
 
@@ -365,16 +411,51 @@ struct session *new_session(struct session *ses, char *name, char *arg, int desc
 		}
 	}
 
-	newses->rows          = gts->rows;
-	newses->cols          = gts->cols;
-	newses->top_row       = gts->top_row;
-	newses->bot_row       = gts->bot_row;
+	newses->event_flags   = gts->event_flags;
 
-	init_buffer(newses, gts->scroll_max);
+	newses->split         = calloc(1, sizeof(struct split_data));
+
+	memcpy(newses->split, gts->split, sizeof(struct split_data));
+
+	newses->cur_row       = gts->cur_row;
+	newses->cur_col       = gts->cur_col;
+
+	newses->wrap          = gts->wrap;
+
+        newses->scroll        = calloc(1, sizeof(struct scroll_data));
+	init_buffer(newses, gts->scroll->size);
+
+	newses->input         = calloc(1, sizeof(struct input_data));
+
+	memcpy(newses->input, gts->input, sizeof(struct input_data));
+
+	newses->input->buf    = str_alloc(BUFFER_SIZE);
+	newses->input->tmp    = str_alloc(BUFFER_SIZE);
+/*
+	newses->input->sav_top_row = gts->input->sav_top_row;
+	newses->input->sav_top_col = gts->input->sav_top_col;
+	newses->input->sav_bot_row = gts->input->sav_bot_row;
+	newses->input->sav_bot_col = gts->input->sav_bot_col;
+
+	newses->input->top_row = gts->input->top_row;
+	newses->input->top_col = gts->input->top_col;
+	newses->input->bot_row = gts->input->bot_row;
+	newses->input->bot_col = gts->input->bot_col;
+*/
+	// may need to set additional values.
+
+	newses->input->off = gts->input->off;
+	newses->input->pos = gts->input->pos;
+	newses->input->hid = gts->input->hid;
+
+	memcpy(&newses->cur_terminal, &gts->cur_terminal, sizeof(gts->cur_terminal));
 
 	if (desc == 0)
 	{
-		tintin_printf(ses, "#TRYING TO CONNECT '%s' TO '%s' PORT '%s'.", newses->name, newses->session_host, newses->session_port);
+		if (!check_all_events(newses, SUB_ARG, 0, 4, "GAG SESSION CREATED", newses->name, newses->session_host, newses->session_ip, newses->session_port))
+		{
+			tintin_printf(ses, "#TRYING TO CONNECT '%s' TO '%s' PORT '%s'.", newses->name, newses->session_host, newses->session_port);
+		}
 	}
 	else if (desc == -1)
 	{
@@ -382,13 +463,19 @@ struct session *new_session(struct session *ses, char *name, char *arg, int desc
 	}
 	else
 	{
-		tintin_printf(ses, "#TRYING TO LAUNCH '%s' RUNNING '%s'.", newses->name, newses->session_host);
+		if (!check_all_events(newses, SUB_ARG, 0, 4, "GAG SESSION CREATED", newses->name, newses->session_host, newses->session_ip, newses->session_port))
+		{
+			tintin_printf(ses, "#TRYING TO LAUNCH '%s' RUNNING '%s'.", newses->name, newses->session_host);
+		}
 	}
 
-	gtd->ses = newses;
-
 	dirty_screen(newses);
 
+	if (gtd->level->background == 0)
+	{
+		gtd->ses = newses;
+	}
+
 	if (desc == 0)
 	{
 		newses = connect_session(newses);
@@ -407,34 +494,40 @@ struct session *new_session(struct session *ses, char *name, char *arg, int desc
 		newses->socket = desc;
 	}
 
-	if (newses)
+	if (newses == NULL)
 	{
-#ifdef HAVE_GNUTLS_H
-
-		if (ssl)
-		{
-			newses->ssl = ssl_negotiate(newses);
-
-			if (newses->ssl == 0)
-			{
-				cleanup_session(newses);
-
-				pop_call();
-				return gtd->ses;
-			}
-		}
-#endif
-		gtd->ses = newses;
-
-		if (*file)
-		{
-			do_read(newses, file);
-		}
-		check_all_events(newses, SUB_ARG, 0, 4, "SESSION CREATED", newses->name, newses->session_host, newses->session_ip, newses->session_port);
+		pop_call();
+		return ses;
 	}
 
+#ifdef HAVE_GNUTLS_H
+	if (ssl)
+	{
+		newses->ssl = ssl_negotiate(newses);
+
+		if (newses->ssl == 0)
+		{
+			cleanup_session(newses);
+
+			pop_call();
+			return ses;
+		}
+	}
+#endif
+
+	if (*file)
+	{
+		newses = execute(newses, "#READ %s", file);
+	}
+	check_all_events(newses, SUB_ARG, 0, 4, "SESSION CREATED", newses->name, newses->session_host, newses->session_ip, newses->session_port);
+
+	if (gtd->level->background == 0)
+	{
+		pop_call();
+		return newses;
+	}
 	pop_call();
-	return gtd->ses;
+	return ses;
 }
 
 struct session *connect_session(struct session *ses)
@@ -452,6 +545,8 @@ struct session *connect_session(struct session *ses)
 
 	if (sock == -1)
 	{
+//		syserr_printf(ses, "connect_session: connect");
+
 		cleanup_session(ses);
 
 		pop_call();
@@ -460,15 +555,22 @@ struct session *connect_session(struct session *ses)
 
 	if (sock)
 	{
+/*
+		if (fcntl(sock, F_SETFL, O_NDELAY|O_NONBLOCK) == -1)
+		{
+			syserr_printf(ses, "connect_session: fcntl O_NDELAY|O_NONBLOCK");
+		}
+*/
 		ses->socket = sock;
 
 		ses->connect_retry = 0;
 
 		SET_BIT(ses->flags, SES_FLAG_CONNECTED);
 
-		tintin_printf2(ses, "");
-
-		tintin_printf(ses, "#SESSION '%s' CONNECTED TO '%s' PORT '%s'", ses->name, ses->session_host, ses->session_port);
+		if (!check_all_events(ses, SUB_ARG, 0, 4, "GAG SESSION CONNECTED", ses->name, ses->session_host, ses->session_ip, ses->session_port))
+		{
+			tintin_printf(ses, "\n#SESSION '%s' CONNECTED TO '%s' PORT '%s'", ses->name, ses->session_host, ses->session_port);
+		}
 
 		check_all_events(ses, SUB_ARG, 0, 4, "SESSION CONNECTED", ses->name, ses->session_host, ses->session_ip, ses->session_port);
 
@@ -531,22 +633,33 @@ void cleanup_session(struct session *ses)
 		gtd->update = ses->next;
 	}
 
+	if (ses == gtd->all)
+	{
+		gtd->all = ses->next;
+	}
+
 	UNLINK(ses, gts->next, gts->prev);
 
 	if (ses->socket)
 	{
 		if (close(ses->socket) == -1)
 		{
-			syserr("close in cleanup");
+			syserr_printf(ses, "cleanup_session: close");
 		}
+//		else
+//		{
+//			int status;
+
+//			wait(&status);
+//		}
 
 		// the PID is stored in the session's port.
-
+/*
 		if (HAS_BIT(ses->flags, SES_FLAG_RUN))
 		{
-			kill(atoi(ses->session_port), SIGKILL);
+			kill(atoi(ses->session_port), SIGTERM);
 		}
-
+*/
 	}
 
 	if (ses->port)
@@ -556,19 +669,30 @@ void cleanup_session(struct session *ses)
 
 	SET_BIT(ses->flags, SES_FLAG_CLOSED);
 
+	client_end_mccp2(ses);
+	client_end_mccp3(ses);
+
 	if (HAS_BIT(ses->flags, SES_FLAG_PORT) || HAS_BIT(ses->flags, SES_FLAG_CONNECTED))
 	{
 		DEL_BIT(ses->flags, SES_FLAG_CONNECTED);
 
+		if (!check_all_events(ses, SUB_ARG, 0, 4, "GAG SESSION DISCONNECTED", ses->name, ses->session_host, ses->session_ip, ses->session_port))
+		{
+			tintin_printf(gtd->ses, "#SESSION '%s' DIED.", ses->name);
+		}
+
 		check_all_events(ses, SUB_ARG, 0, 4, "SESSION DISCONNECTED", ses->name, ses->session_host, ses->session_ip, ses->session_port);
 
-		tintin_printf(gtd->ses, "#SESSION '%s' DIED.", ses->name);
+
 	}
 	else
 	{
-		check_all_events(ses, SUB_ARG, 0, 4, "SESSION TIMED OUT", ses->name, ses->session_host, ses->session_ip, ses->session_port);
+		if (!check_all_events(ses, SUB_ARG, 0, 4, "GAG SESSION TIMED OUT", ses->name, ses->session_host, ses->session_ip, ses->session_port))
+		{
+			tintin_printf(gtd->ses, "#SESSION '%s' TIMED OUT.", ses->name);
+		}
 
-		tintin_printf(gtd->ses, "#SESSION '%s' TIMED OUT.", ses->name);
+		check_all_events(ses, SUB_ARG, 0, 4, "SESSION TIMED OUT", ses->name, ses->session_host, ses->session_ip, ses->session_port);
 	}
 
 	if (ses == gtd->ses)
@@ -576,21 +700,6 @@ void cleanup_session(struct session *ses)
 		gtd->ses = newactive_session();
 	}
 
-	if (ses->logfile)
-	{
-		fclose(ses->logfile);
-	}
-
-	if (ses->lognext_file)
-	{
-		fclose(ses->lognext_file);
-	}
-
-	if (ses->logline_file)
-	{
-		fclose(ses->logline_file);
-	}
-
 #ifdef HAVE_GNUTLS_H
 
 	if (ses->ssl)
@@ -614,9 +723,19 @@ void dispose_session(struct session *ses)
 
 	UNLINK(ses, gtd->dispose_next, gtd->dispose_prev);
 
-	for (index = 0 ; index < LIST_MAX ; index++)
+	if (ses->logfile)
 	{
-		free_list(ses->list[index]);
+		fclose(ses->logfile);
+	}
+
+	if (ses->lognext_file)
+	{
+		fclose(ses->lognext_file);
+	}
+
+	if (ses->logline_file)
+	{
+		fclose(ses->logline_file);
 	}
 
 	if (ses->map)
@@ -624,10 +743,9 @@ void dispose_session(struct session *ses)
 		delete_map(ses);
 	}
 
-	if (ses->mccp)
+	for (index = 0 ; index < LIST_MAX ; index++)
 	{
-		inflateEnd(ses->mccp);
-		free(ses->mccp);
+		free_list(ses->list[index]);
 	}
 
 	init_buffer(ses, 0);
@@ -639,7 +757,13 @@ void dispose_session(struct session *ses)
 	free(ses->group);
 	free(ses->read_buf);
 	free(ses->cmd_color);
+	free(ses->lognext_name);
 	free(ses->logline_name);
+	free(ses->split);
+	str_free(ses->input->buf);
+	str_free(ses->input->tmp);
+	free(ses->input);
+
 	free(ses);
 
 	pop_call();
