@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2008-2019 Adam Borowski and Igor van den Hoven                  *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,16 +13,16 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
 ******************************************************************************/
 
 /******************************************************************************
 *                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
 *                                                                             *
 *                         coded by Adam Borowski 2008                         *
-*                       added by Igor van den Hoven 2014                      *
+*                   modifications by Igor van den Hoven 2014                  *
 ******************************************************************************/
 
 #include "tintin.h"
@@ -36,23 +35,23 @@ static int ssl_check_cert(struct session *ses, gnutls_session_t sslses);
 
 DO_COMMAND(do_ssl)
 {
-	char temp[BUFFER_SIZE], left[BUFFER_SIZE];
+	char temp[BUFFER_SIZE];
 
 	substitute(ses, arg, temp, SUB_VAR|SUB_FUN);
 
 	arg = temp;
 
-	arg = get_arg_in_braces(ses, arg, left, FALSE);
+	arg = get_arg_in_braces(ses, arg, arg1, GET_ONE);
 
-	if (*left == 0 || *arg == 0)
+	if (*arg1 == 0 || *arg == 0)
 	{
-		tintin_printf2(ses, "#SYNTAX: #SSL {name} {host} {port}");
+		show_error(ses, LIST_COMMAND, "#SYNTAX: #SSL {name} {host} {port}");
 	}
 	else
 	{
-		ses = new_session(ses, left, arg, 0, 1);
+		ses = new_session(ses, arg1, arg, 0, 1);
 	}
-	return gtd->ses;
+	return ses;
 }
 
 gnutls_session_t ssl_negotiate(struct session *ses)
@@ -70,7 +69,7 @@ gnutls_session_t ssl_negotiate(struct session *ses)
 	gnutls_init(&ssl_ses, GNUTLS_CLIENT);
 	gnutls_set_default_priority(ssl_ses);
 	gnutls_credentials_set(ssl_ses, GNUTLS_CRD_CERTIFICATE, ssl_cred);
-	gnutls_transport_set_ptr(ssl_ses, (gnutls_transport_ptr_t) ses->socket);
+	gnutls_transport_set_ptr(ssl_ses, (gnutls_transport_ptr_t) (long int) ses->socket);
 
 	do 
 	{
@@ -80,7 +79,7 @@ gnutls_session_t ssl_negotiate(struct session *ses)
 
 	if (ret)
 	{
-		tintin_printf2(ses, "#SSL handshake failed: %s", gnutls_strerror(ret));
+		tintin_printf2(ses, "#SSL: handshake failed error: %s", gnutls_strerror(ret));
 		gnutls_deinit(ssl_ses);
 		return 0;
 	}
