@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2019 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -14,9 +13,9 @@
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *   GNU General Public License for more details.                              *
 *                                                                             *
+*                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
 ******************************************************************************/
 
 /******************************************************************************
@@ -34,132 +33,63 @@
 #include <util.h>
 #endif
 #endif
+#include <fcntl.h>  
+#include <dirent.h>
+#include <termios.h>
+#include <sys/un.h>
+
+#include <sys/wait.h>
 
 DO_COMMAND(do_run)
 {
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], temp[BUFFER_SIZE], file[BUFFER_SIZE];
+	char temp[BUFFER_SIZE];
 	int desc, pid;
 	struct winsize size;
 
 	char *argv[4] = {"sh", "-c", "", NULL};
 
-	arg = sub_arg_in_braces(ses, arg, left,  GET_ONE, SUB_VAR|SUB_FUN);
-	arg = sub_arg_in_braces(ses, arg, right, GET_ALL, SUB_VAR|SUB_FUN);
-	arg = sub_arg_in_braces(ses, arg, file,  GET_ONE, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg2, GET_ALL, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg3, GET_ONE, SUB_VAR|SUB_FUN);
 
-	if (*left == 0 || *right == 0)
+	if (*arg1 == 0 || *arg2 == 0)
 	{
-		return show_error(ses, LIST_COMMAND, "#SYNTAX: #RUN {NAME} {SYSTEM SHELL COMMAND}");
+		show_error(ses, LIST_COMMAND, "#SYNTAX: #RUN {NAME} {SYSTEM SHELL COMMAND}");
+		
+		return ses;
 	}
 
-	size.ws_row = get_scroll_size(ses);
-	size.ws_col = ses->cols;
+	size.ws_row = get_scroll_rows(ses);
+	size.ws_col = get_scroll_cols(ses);
+	size.ws_ypixel = size.ws_row * gtd->screen->char_height;
+	size.ws_xpixel = size.ws_col * gtd->screen->char_width;
 
 	pid = forkpty(&desc, temp, &gtd->old_terminal, &size);
-//	pid = forkpty(&desc, temp, &gtd->new_terminal, &size);
 
 	switch (pid)
 	{
 		case -1:
-			perror("forkpty");
+			syserr_printf(ses, "do_run: forkpty");
 			break;
 
 		case 0:
-			sprintf(temp, "exec %s", right);
+			sprintf(temp, "exec %s", arg2);
 			argv[2] = temp;
 			execv("/bin/sh", argv);
 			break;
 
 		default:
-			sprintf(temp, "{%s} {%d} {%s}", right, pid, file);
-
-			ses = new_session(ses, left, temp, desc, 0);
-
+			sprintf(temp, "{%s} {%d} {%s}", arg2, pid, arg3);
+			ses = new_session(ses, arg1, temp, desc, 0);
 			break;
 	}
-	return gtd->ses;
-}
-
-DO_COMMAND(do_scan)
-{
-	FILE *fp;
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], filename[BUFFER_SIZE];
-
-	arg = sub_arg_in_braces(ses, arg, left, GET_ONE, SUB_VAR|SUB_FUN);
-	arg = sub_arg_in_braces(ses, arg, right, GET_ONE, SUB_VAR|SUB_FUN);
-
-	if (*left == 0)
-	{
-		show_error(ses, LIST_COMMAND, "#SYNTAX: #SCAN {ABORT|READ} {<FILENAME>}");
-		
-		return ses;
-	}
-
-	if (is_abbrev(left, "ABORT"))
-	{
-		if (!HAS_BIT(ses->flags, SES_FLAG_SCAN))
-		{
-			show_error(ses, LIST_COMMAND, "#SCAN ABORT: NOT CURRENTLY SCANNING.");
-		}
-		else
-		{
-			SET_BIT(ses->flags, SES_FLAG_SCANABORT);
-		}
-		return ses;
-	}
-
-	if (is_abbrev(left, "READ") && *right)
-	{
-		strcpy(left, right);
-	}
-
-	if ((fp = fopen(left, "r")) == NULL)
-	{
-		show_error(ses, LIST_COMMAND, "#ERROR: #SCAN READ {%s} - FILE NOT FOUND.", left);
-
-		return ses;
-	}
-
-	SET_BIT(ses->flags, SES_FLAG_SCAN);
-
-	if (STRING_SIZE > gtd->mud_output_max)
-	{
-		gtd->mud_output_max  = STRING_SIZE;
-		gtd->mud_output_buf  = (char *) realloc(gtd->mud_output_buf, gtd->mud_output_max);
-	}
-
-	while (fgets(gtd->mud_output_buf, STRING_SIZE, fp))
-	{
-		gtd->mud_output_len = 1;
-
-		readmud(ses);
-
-		if (HAS_BIT(ses->flags, SES_FLAG_SCANABORT))
-		{
-			break;
-		}
-	}
-
-	DEL_BIT(ses->flags, SES_FLAG_SCAN);
-
-	if (HAS_BIT(ses->flags, SES_FLAG_SCANABORT))
-	{
-		DEL_BIT(ses->flags, SES_FLAG_SCANABORT);
-
-		show_message(ses, LIST_COMMAND, "#SCAN: FILE PARTIALLY SCANNED.", filename);
-	}
-	else
-	{
-		show_message(ses, LIST_COMMAND, "#SCAN: FILE SCANNED.", filename);
-	}
-	fclose(fp);
-
 	return ses;
 }
 
+
 DO_COMMAND(do_script)
 {
-	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], *cptr, buf[BUFFER_SIZE], var[BUFFER_SIZE], tmp[BUFFER_SIZE];
+	char *cptr, buf[BUFFER_SIZE], var[BUFFER_SIZE], tmp[BUFFER_SIZE];
 	FILE *script;
 	int index;
 
@@ -192,7 +122,7 @@ DO_COMMAND(do_script)
 		}
 		else
 		{
-			perror("popen");
+			syserr_printf(ses, "do_script: popen 1");
 		}
 	}
 	else
@@ -219,17 +149,16 @@ DO_COMMAND(do_script)
 				cat_sprintf(var, "{%d}{%s}", index++, tmp);
 			}
 
-
-			set_nest_node(ses->list[LIST_VARIABLE], arg1, "%s", var);
+			set_nest_node_ses(ses, arg1, "%s", var);
 
 			pclose(script);
 		}
 		else
 		{
-			perror("popen");
+			syserr_printf(ses, "do_script: popen 2");
 		}
 	}
-	refresh_terminal();
+	refresh_session_terminal(ses);
 
 	return ses;
 }
@@ -237,42 +166,138 @@ DO_COMMAND(do_script)
 
 DO_COMMAND(do_suspend)
 {
-	suspend_handler(0);
+	print_stdout("\e[r\e[%d;%dH", gtd->screen->rows, 1);
+
+	fflush(NULL);
+
+	reset_terminal(gtd->ses);
+
+	kill(0, SIGSTOP);
+
+	init_terminal(gtd->ses);
+
+	dirty_screen(gtd->ses);
+
+	tintin_puts(NULL, "#RETURNING BACK TO TINTIN++.");
 
 	return ses;
 }
 
+// use #run + waitpid() ?
+
+/*
+DO_COMMAND(do_system)
+{
+	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], temp[BUFFER_SIZE], buffer[BUFFER_SIZE];
+	int desc, pid;
+	struct winsize size;
+	char *argv[4] = {"sh", "-c", "", NULL};
+	int filedes[2];
+
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
+
+	if (*arg1 == 0)
+	{
+		show_error(ses, LIST_COMMAND, "#SYNTAX: #SYSTEM {SHELL COMMAND}");
+		
+		return ses;
+	}
+
+	size.ws_row = get_scroll_rows(ses);
+	size.ws_col = get_scroll_cols(ses);
+	size.ws_ypixel = size.ws_row * gtd->screen->char_height;
+	size.ws_xpixel = size.ws_col * gtd->screen->char_width;
+
+	if (pipe(filedes) == -1)
+	{
+		perror("pipe");
+
+		return ses;
+	}
+
+	pid = fork();
+
+	switch (pid)
+	{
+		case -1:
+			syserr_printf(ses, "do_system: forkpty");
+			break;
+
+		case 0:
+			while ((dup2(filedes[1], STDOUT_FILENO) == -1) && (errno == EINTR)) {}
+			close(filedes[1]);
+			close(filedes[0]);
+
+			sprintf(temp, "exec %s", arg1);
+			argv[2] = temp;
+			execv("/bin/sh", argv);
+			perror("execv");
+			break;
+
+		default:
+			close(filedes[1]);
+
+			while (TRUE)
+			{
+				ssize_t count = read(filedes[0], buffer, sizeof(buffer));
+
+				if (count == -1)
+				{
+					if (errno == EINTR)
+					{
+						continue;
+					}
+					else
+					{
+						perror("read");
+					}
+					return ses;
+				}
+				else if (count == 0)
+				{
+					break;
+				}
+				else
+				{
+					tintin_printf2(ses, "%s", buffer);
+				}
+			}
+			close(filedes[0]);
+			wait(0);
+			break;
+	}
+	return ses;
+}
+*/
 
 DO_COMMAND(do_system)
 {
-	char left[BUFFER_SIZE];
+	sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
 
-	get_arg_in_braces(ses, arg, left, TRUE);
-	substitute(ses, left, left, SUB_VAR|SUB_FUN);
-
-	if (*left == 0)
+	if (*arg1 == 0)
 	{
-		return show_error(ses, LIST_COMMAND, "#SYNTAX: #SYSTEM {COMMAND}.");
+		show_error(ses, LIST_COMMAND, "#SYNTAX: #SYSTEM {COMMAND}.");
+		
+		return ses;
 	}
 
-	show_message(ses, LIST_COMMAND, "#OK: EXECUTING '%s'", left);
+	show_message(ses, LIST_COMMAND, "#OK: EXECUTING '%s'", arg1);
 
 	if (!HAS_BIT(gtd->ses->flags, SES_FLAG_READMUD) && IS_SPLIT(gtd->ses))
 	{
 		save_pos(gtd->ses);
-		goto_rowcol(gtd->ses, gtd->ses->bot_row, 1);
-	}
-	fflush(stdout);
 
-	system(left);
+		goto_pos(gtd->ses, gtd->ses->split->bot_row, 1);
+	}
+
+	system(arg1);
 
 	if (!HAS_BIT(gtd->ses->flags, SES_FLAG_READMUD) && IS_SPLIT(gtd->ses))
 	{
 		restore_pos(gtd->ses);
 	}
-	fflush(stdout);
 
-	refresh_terminal();
+	refresh_session_terminal(gtd->ses);
 
 	return ses;
 }
@@ -281,16 +306,16 @@ DO_COMMAND(do_system)
 DO_COMMAND(do_textin)
 {
 	FILE *fp;
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], buffer[BUFFER_SIZE], *cptr;
+	char buffer[BUFFER_SIZE], *cptr;
 
-	arg = get_arg_in_braces(ses, arg, left, FALSE);
-	substitute(ses, left, left, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ONE, SUB_VAR|SUB_FUN);
+	arg = get_arg_in_braces(ses, arg, arg2, GET_ALL);
 
-	arg = get_arg_in_braces(ses, arg, right, TRUE);
-
-	if ((fp = fopen(left, "r")) == NULL)
+	if ((fp = fopen(arg1, "r")) == NULL)
 	{
-		return show_error(ses, LIST_COMMAND, "#ERROR: #TEXTIN {%s} - FILE NOT FOUND.", left);
+		show_error(ses, LIST_COMMAND, "#ERROR: #TEXTIN {%s} - FILE NOT FOUND.", arg1);
+		
+		return ses;
 	}
 
 	while (fgets(buffer, BUFFER_SIZE - 1, fp))
@@ -304,14 +329,14 @@ DO_COMMAND(do_textin)
 
 		write_mud(ses, buffer, SUB_EOL);
 
-		if (*right)
+		if (*arg2)
 		{
-			usleep((long long) (get_number(ses, right) * 1000000));
+			usleep((long long) (get_number(ses, arg2) * 1000000));
 		}
 	}
 	fclose(fp);
 
-	show_message(ses, LIST_COMMAND, "#OK. FILE READ.");
+	show_message(ses, LIST_COMMAND, "#TEXTIN {%s} - FILE READ.", arg1);
 
 	return ses;
 }
