@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2020 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -15,14 +14,13 @@
 *   GNU General Public License for more details.                              *
 *                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
 ******************************************************************************/
 
 /******************************************************************************
-*               (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                  *
+*                               T I N T I N + +                               *
 *                                                                             *
-*                     coded by Igor van den Hoven 2008                        *
+*                      coded by Igor van den Hoven 2008                       *
 ******************************************************************************/
 
 #include "tintin.h"
@@ -35,6 +33,7 @@ struct scriptdata
 	long long              cnt;
 	int                    inc;
 	char                 * cpy;
+	char                 * hlt;
 	char                 * str;
 	char                 * arg;
 };
@@ -62,67 +61,57 @@ struct script_regex
 	int                    val;
 };
 
-struct scriptroot
-{
-	struct scriptnode    * next;
-	struct scriptnode    * prev;
-	struct session       * ses;
-	struct listroot      * local;
-	int list;
-};
-
-struct scriptroot *script_stack[1001];
-
-int script_index;
-
 void debugtoken(struct session *ses, struct scriptroot *root, struct scriptnode *token)
 {
 	push_call("debugtoken(%p,%d,%p,%d)",ses,root->list,token,token->type);
 
-	if (gtd->debug_level)
+	if (gtd->level->debug || HAS_BIT(root->ses->list[root->list]->flags, LIST_FLAG_DEBUG) || HAS_BIT(root->ses->list[root->list]->flags, LIST_FLAG_LOG))
 	{
 		switch (token->type)
 		{
 			case TOKEN_TYPE_STRING:
 			case TOKEN_TYPE_SESSION:
-				show_debug(ses, root->list, "[%02d] %*s%s", token->type, token->lvl * 4, "", token->str);
+				show_debug(ses, root->list, "[%02d] %s%s", token->type, indent(token->lvl), token->str);
 				break;
 
 			case TOKEN_TYPE_ELSE:
 			case TOKEN_TYPE_END:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;32m%s\e[0m", token->type, token->lvl * 4, "", token->str);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s\e[0m", token->type, indent(token->lvl), token->str);
 				break;
 
 			case TOKEN_TYPE_DEFAULT:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;32m%s\e[0m", token->type, token->lvl * 4, "", command_table[token->cmd].name);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name);
 				break;
 
 			case TOKEN_TYPE_BREAK:
 			case TOKEN_TYPE_CONTINUE:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;31m%s\e[0m", token->type, token->lvl * 4, "", command_table[token->cmd].name);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name);
 				break;
 
 			case TOKEN_TYPE_COMMAND:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;36m%s\e[0m %s\e[0m", token->type, token->lvl * 4, "", command_table[token->cmd].name, token->str);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_COMMAND   "%s " COLOR_STRING "%s\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name, token->str);
 				break;
 
 			case TOKEN_TYPE_RETURN:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;31m%s\e[0m %s\e[0m", token->type, token->lvl * 4, "", command_table[token->cmd].name, token->str);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s " COLOR_STRING "%s\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name, token->str);
 				break;
 
 			case TOKEN_TYPE_CASE:
 			case TOKEN_TYPE_ELSEIF:
 			case TOKEN_TYPE_IF:
+			case TOKEN_TYPE_WHILE:
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s " COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "}\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name, token->str);
+				break;
+
 			case TOKEN_TYPE_FOREACH:
 			case TOKEN_TYPE_LOOP:
 			case TOKEN_TYPE_PARSE:
 			case TOKEN_TYPE_SWITCH:
-			case TOKEN_TYPE_WHILE:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;32m%s {\e[0m%s\e[1;32m}\e[0m", token->type, token->lvl * 4, "", command_table[token->cmd].name, token->str);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s " COLOR_STRING "%s\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name, token->data->hlt);
 				break;
 
 			case TOKEN_TYPE_REGEX:
-				show_debug(ses, root->list, "[%02d] %*s\e[1;32m%s {\e[0m%s\e[1;32m} {\e[0m%s\e[1;32m}\e[0m", token->type, token->lvl * 4, "", command_table[token->cmd].name, token->str, token->regex->str);
+				show_debug(ses, root->list, "[%02d] %s" COLOR_STATEMENT "%s " COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "} {" COLOR_STRING "%s" COLOR_BRACE "}\e[0m", token->type, indent(token->lvl), command_table[token->cmd].name, token->str, token->regex->str);
 				break;
 
 			default:
@@ -132,7 +121,7 @@ void debugtoken(struct session *ses, struct scriptroot *root, struct scriptnode
 				}
 				else
 				{
-					show_debug(ses, root->list, "[%02d] %*s\e[1;33m%d {\e[0m%s\e[1;32m}\e[0m", token->type, token->lvl * 4, "", token->cmd, token->str);
+					show_debug(ses, root->list, "[%02d] %s\e[1;33m%d {\e[0m%s\e[1;32m}\e[0m", token->type, indent(token->lvl), token->cmd, token->str);
 				}
 				break;
 		}
@@ -165,11 +154,12 @@ char *addlooptoken(struct scriptroot *root, int lvl, int opr, int cmd, char *str
 
 	data = (struct scriptdata *) calloc(1, sizeof(struct scriptdata));
 
-	str = get_arg_in_braces(root->ses, str, min, FALSE);
-	str = get_arg_in_braces(root->ses, str, max, FALSE);
-	str = get_arg_in_braces(root->ses, str, var, FALSE);
+	str = get_arg_in_braces(root->ses, str, min, GET_ONE);
+	str = get_arg_in_braces(root->ses, str, max, GET_ONE);
+	str = get_arg_in_braces(root->ses, str, var, GET_ONE);
 
-	data->cpy = refstring(NULL, "{%s} {%s} {%s}", min, max, var);
+	data->cpy = restringf(NULL, "{%s} {%s} {%s}", min, max, var);
+	data->hlt = restringf(NULL, COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "} {" COLOR_STRING "%s" COLOR_BRACE "} {" COLOR_STRING "%s" COLOR_BRACE "}", min, max, var);
 
 	addtoken(root, lvl, opr, cmd, var);
 
@@ -188,9 +178,10 @@ char *addswitchtoken(struct scriptroot *root, int lvl, int opr, int cmd, char *s
 
 	data = (struct scriptdata *) calloc(1, sizeof(struct scriptdata));
 
-	str = get_arg_in_braces(root->ses, str, arg, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg, GET_ONE);
 
-	data->cpy = strdup("");
+	data->cpy = restringf(NULL, "{%s}", arg);
+	data->hlt = restringf(NULL, COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "}", arg);
 
 	addtoken(root, lvl, opr, cmd, arg);
 
@@ -207,8 +198,8 @@ void resetlooptoken(struct session *ses, struct scriptnode *token)
 
 	str = token->data->cpy;
 
-	str = get_arg_in_braces(ses, str, min, FALSE);
-	str = get_arg_in_braces(ses, str, max, FALSE);
+	str = get_arg_in_braces(ses, str, min, GET_ONE);
+	str = get_arg_in_braces(ses, str, max, GET_ONE);
 
 	token->data->min = (int) get_number(ses, min);
 	token->data->max = (int) get_number(ses, max);
@@ -228,14 +219,15 @@ char *addforeachtoken(struct scriptroot *root, int lvl, int opr, int cmd, char *
 
 	char arg[BUFFER_SIZE], var[BUFFER_SIZE];
 
-	str = get_arg_in_braces(root->ses, str, arg, FALSE);
-	str = get_arg_in_braces(root->ses, str, var, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg, GET_ONE);
+	str = get_arg_in_braces(root->ses, str, var, GET_ONE);
 
 	addtoken(root, lvl, opr, cmd, var);
 
 	data = (struct scriptdata *) calloc(1, sizeof(struct scriptdata));
 
-	data->cpy = refstring(NULL, "{%s} {%s}", arg, var);
+	data->cpy = restringf(NULL, "{%s} {%s}", arg, var);
+	data->hlt = restringf(NULL, COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "} {" COLOR_STRING "%s" COLOR_BRACE "}", arg, var);
 	data->str = strdup("");
 	data->arg = data->str;
 
@@ -270,7 +262,7 @@ void handlereturntoken(struct session *ses, struct scriptnode *token)
 
 	substitute(ses, token->str, arg, SUB_VAR|SUB_FUN);
 
-	set_nest_node(ses->list[LIST_VARIABLE], "result", "%s", arg);
+	set_nest_node_ses(ses, "result", "%s", arg);
 }
 
 void handleswitchtoken(struct session *ses, struct scriptnode *token)
@@ -286,7 +278,7 @@ char *get_arg_foreach(struct scriptroot *root, struct scriptnode *token)
 {
 	static char buf[BUFFER_SIZE];
 
-	token->data->arg = get_arg_in_braces(root->ses, token->data->arg, buf, TRUE);
+	token->data->arg = get_arg_in_braces(root->ses, token->data->arg, buf, GET_ALL);
 
 	if (*token->data->arg == COMMAND_SEPARATOR)
 	{
@@ -301,14 +293,15 @@ char *addparsetoken(struct scriptroot *root, int lvl, int opr, int cmd, char *st
 
 	char arg[BUFFER_SIZE], var[BUFFER_SIZE];
 
-	str = get_arg_in_braces(root->ses, str, arg, FALSE);
-	str = get_arg_in_braces(root->ses, str, var, FALSE);
+	str = get_arg_in_braces(root->ses, str, arg, GET_ONE);
+	str = get_arg_in_braces(root->ses, str, var, GET_ONE);
 
 	addtoken(root, lvl, opr, cmd, var);
 
 	data = (struct scriptdata *) calloc(1, sizeof(struct scriptdata));
 
-	data->cpy = refstring(NULL, "{%s} {%s}", arg, var);
+	data->cpy = restringf(NULL, "{%s} {%s}", arg, var);
+	data->hlt = restringf(NULL, COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "} {" COLOR_STRING "%s" COLOR_BRACE "}", arg, var);
 
 	data->str = strdup("");
 	data->arg = data->str;
@@ -342,24 +335,13 @@ char *get_arg_parse(struct session *ses, struct scriptnode *token)
 {
 	static char buf[5];
 
-	if (HAS_BIT(ses->flags, SES_FLAG_BIG5) && token->data->arg[0] & 128 && token->data->arg[1] != 0)
+	if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, token->data->arg))
 	{
-		token->data->arg += sprintf(buf, "%c%c", token->data->arg[0], token->data->arg[1]);
+		token->data->arg += sprintf(buf, "%.*s", get_euc_size(ses, token->data->arg), token->data->arg);
 	}
-	else if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (token->data->arg[0] & 192) == 192 && token->data->arg[1] != 0)
+	else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(token->data->arg))
 	{
-		if ((token->data->arg[0] & 240) == 240 && token->data->arg[2] != 0 && token->data->arg[3] != 0)
-		{
-			token->data->arg += sprintf(buf, "%c%c%c%c", token->data->arg[0], token->data->arg[1], token->data->arg[2], token->data->arg[3]);
-		}
-		else if ((token->data->arg[0] & 224) == 224 && token->data->arg[2] != 0)
-		{
-			token->data->arg += sprintf(buf, "%c%c%c", token->data->arg[0], token->data->arg[1], token->data->arg[2]);
-		}
-		else
-		{
-			token->data->arg += sprintf(buf, "%c%c", token->data->arg[0], token->data->arg[1]);
-		}
+		token->data->arg += sprintf(buf, "%.*s", get_utf8_size(token->data->arg), token->data->arg);
 	}
 	else
 	{
@@ -375,9 +357,9 @@ char *addregextoken(struct scriptroot *root, int lvl, int type, int cmd, char *s
 
 	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], arg3[BUFFER_SIZE];
 
-	str = get_arg_in_braces(root->ses, str, arg1, FALSE);
-	str = get_arg_in_braces(root->ses, str, arg2, FALSE);
-	str = get_arg_in_braces(root->ses, str, arg3, TRUE);
+	str = get_arg_in_braces(root->ses, str, arg1, GET_ONE);
+	str = get_arg_in_braces(root->ses, str, arg2, GET_ONE);
+	str = get_arg_in_braces(root->ses, str, arg3, GET_ALL);
 
 	addtoken(root, lvl, type, cmd, arg1);
 
@@ -414,6 +396,7 @@ void deltoken(struct scriptroot *root, struct scriptnode *token)
 		case TOKEN_TYPE_PARSE:
 		case TOKEN_TYPE_SWITCH:
 			free(token->data->cpy);
+			free(token->data->hlt);
 			free(token->data->str);
 			free(token->data);
 			break;
@@ -435,7 +418,7 @@ int find_command(char *command)
 		return -1;
 	}
 
-	if (isalpha((int) *command))
+	if (isalpha((int) *command) && command[1] != 0)
 	{
 		for (cmd = gtd->command_ref[tolower((int) *command) - 'a'] ; *command_table[cmd].name ; cmd++)
 		{
@@ -458,7 +441,7 @@ void init_local(struct session *ses)
 	root->list = LIST_VARIABLE;
 	root->local = init_list(ses, LIST_VARIABLE, LIST_SIZE);
 
-	script_stack[0] = root;
+	gtd->script_stack[0] = root;
 
 	return;
 }
@@ -469,7 +452,7 @@ struct listroot *local_list(struct session *ses)
 
 	push_call("local_list(%p)",ses);
 
-	root = script_stack[script_index]->local;
+	root = gtd->script_stack[gtd->script_index]->local;
 
 	pop_call();
 	return root;
@@ -510,7 +493,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 
 			if (cmd == -1)
 			{
-				str = get_arg_all(root->ses, str, line, FALSE);
+				str = get_arg_all(root->ses, str, line, 0);
 				addtoken(root, lvl, TOKEN_TYPE_SESSION, -1, line+1);
 			}
 			else
@@ -523,10 +506,10 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 						break;
 
 					case TOKEN_TYPE_CASE:
-						str = get_arg_in_braces(root->ses, arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ONE);
 						addtoken(root, lvl++, TOKEN_TYPE_CASE, cmd, line);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endcase");
@@ -540,7 +523,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_DEFAULT:
 						addtoken(root, lvl++, TOKEN_TYPE_DEFAULT, cmd, "");
 
-						str = get_arg_in_braces(root->ses, arg, line, TRUE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "enddefault");
@@ -549,17 +532,17 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_ELSE:
 						addtoken(root, lvl++, TOKEN_TYPE_ELSE, cmd, "else");
 
-						str = get_arg_in_braces(root->ses, arg, line, TRUE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endelse");
 						break;
 
 					case TOKEN_TYPE_ELSEIF:
-						str = get_arg_in_braces(root->ses, arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ONE);
 						addtoken(root, lvl++, TOKEN_TYPE_ELSEIF, cmd, line);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endif");
@@ -568,17 +551,17 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_FOREACH:
 						str = addforeachtoken(root, lvl++, TOKEN_TYPE_FOREACH, cmd, arg);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endforeach");
 						break;
 
 					case TOKEN_TYPE_IF:
-						str = get_arg_in_braces(root->ses, arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ONE);
 						addtoken(root, lvl++, TOKEN_TYPE_IF, cmd, line);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endif");
@@ -587,7 +570,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 						{
 							addtoken(root, lvl++, TOKEN_TYPE_ELSE, -1, "else");
 
-							str = get_arg_in_braces(root->ses, str, line, TRUE);
+							str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 							tokenize_script(root, lvl--, line);
 
 							addtoken(root, lvl, TOKEN_TYPE_END, -1, "endif");
@@ -597,7 +580,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_LOOP:
 						str = addlooptoken(root, lvl++, TOKEN_TYPE_LOOP, cmd, arg);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endloop");
@@ -606,7 +589,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 					case TOKEN_TYPE_PARSE:
 						str = addparsetoken(root, lvl++, TOKEN_TYPE_PARSE, cmd, arg);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endparse");
@@ -621,7 +604,7 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 						{
 							addtoken(root, lvl++, TOKEN_TYPE_ELSE, -1, "else");
 
-							str = get_arg_in_braces(root->ses, str, line, TRUE);
+							str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 							tokenize_script(root, lvl--, line);
 
 							addtoken(root, lvl, TOKEN_TYPE_END, -1, "endregex");
@@ -629,31 +612,31 @@ void tokenize_script(struct scriptroot *root, int lvl, char *str)
 						break;
 
 					case TOKEN_TYPE_RETURN:
-						str = get_arg_in_braces(root->ses, arg, line, TRUE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ALL);
 						addtoken(root, lvl, TOKEN_TYPE_RETURN, cmd, line);
 						break;
 
 					case TOKEN_TYPE_SWITCH:
 						str = addswitchtoken(root, lvl++, TOKEN_TYPE_SWITCH, cmd, arg);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "end");
 						break;
 
 					case TOKEN_TYPE_WHILE:
-						str = get_arg_in_braces(root->ses, arg, line, FALSE);
+						str = get_arg_in_braces(root->ses, arg, line, GET_ONE);
 						addtoken(root, lvl++, TOKEN_TYPE_WHILE, cmd, line);
 
-						str = get_arg_in_braces(root->ses, str, line, TRUE);
+						str = get_arg_in_braces(root->ses, str, line, GET_ALL);
 						tokenize_script(root, lvl--, line);
 
 						addtoken(root, lvl, TOKEN_TYPE_END, -1, "endwhile");
 						break;
 
 					default:
-						str = get_arg_with_spaces(root->ses, arg, line, TRUE);
+						str = get_arg_with_spaces(root->ses, arg, line, GET_ALL);
 						addtoken(root, lvl, TOKEN_TYPE_COMMAND, cmd, line);
 						break;
 				}
@@ -760,9 +743,31 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 			case TOKEN_TYPE_COMMAND:
 				push_call("do_%s(%p,%p)", command_table[token->cmd].name, root->ses, token->str);
 
-				root->ses = (*command_table[token->cmd].command) (root->ses, token->str);
+				switch (command_table[token->cmd].args)
+				{
+					case 0:
+						root->ses = (*command_table[token->cmd].command) (root->ses, token->str, NULL, NULL, NULL, NULL);
+						break;
+					case 1:
+						root->ses = (*command_table[token->cmd].command) (root->ses, token->str, str_alloc_stack(), NULL, NULL, NULL);
+						break;
+					case 2:
+						root->ses = (*command_table[token->cmd].command) (root->ses, token->str, str_alloc_stack(), str_alloc_stack(), NULL, NULL);
+						break;
+					case 3:
+						root->ses = (*command_table[token->cmd].command) (root->ses, token->str, str_alloc_stack(), str_alloc_stack(), str_alloc_stack(), NULL);
+						break;
+					case 4:
+						root->ses = (*command_table[token->cmd].command) (root->ses, token->str, str_alloc_stack(), str_alloc_stack(), str_alloc_stack(), str_alloc_stack());
+						break;
+				}
 
 				pop_call();
+/*
+	return;
+}
+*/
+
 				break;
 
 			case TOKEN_TYPE_CONTINUE:
@@ -841,7 +846,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 				}
 				else
 				{
-					set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%s", get_arg_foreach(root, token));
+					set_nest_node_ses(root->ses, token->str, "%s", get_arg_foreach(root, token));
 
 					if (*token->data->arg == 0)
 					{
@@ -876,7 +881,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 					resetlooptoken(root->ses, token);
 				}
 
-				set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%lld", token->data->cnt);
+				set_nest_node_ses(root->ses, token->str, "%lld", token->data->cnt);
 
 				token->data->cnt += token->data->inc;
 
@@ -909,7 +914,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 
 				}
 
-				set_nest_node(root->ses->list[LIST_VARIABLE], token->str, "%s", get_arg_parse(root->ses, token));
+				set_nest_node_ses(root->ses, token->str, "%s", get_arg_parse(root->ses, token));
 
 				if (*token->data->arg == 0)
 				{
@@ -922,7 +927,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 			case TOKEN_TYPE_REGEX:
 				split = NULL;
 
-				token->regex->val = find(root->ses, token->str, token->regex->str, SUB_CMD);
+				token->regex->val = find(root->ses, token->str, token->regex->str, SUB_VAR|SUB_FUN, REGEX_FLAG_CMD);
 
 				if (token->regex->val)
 				{
@@ -970,7 +975,7 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 				}
 				else
 				{
-					token->type++;
+//					token->type++;
 
 					do
 					{
@@ -986,14 +991,13 @@ struct scriptnode *parse_script(struct scriptroot *root, int lvl, struct scriptn
 			token = token->next;
 		}
 	}
+
 	if (lvl)
 	{
 		return NULL;
 	}
-	else
-	{
-		return (struct scriptnode *) root->ses;
-	}
+
+	return (struct scriptnode *) root->ses;
 }
 
 
@@ -1035,19 +1039,19 @@ char *write_script(struct session *ses, struct scriptroot *root)
 			case TOKEN_TYPE_FOREACH:
 			case TOKEN_TYPE_LOOP:
 			case TOKEN_TYPE_PARSE:
+			case TOKEN_TYPE_SWITCH:
 				cat_sprintf(buf, "%s%c%s %s\n%s{\n", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->data->cpy, indent(token->lvl));
 				break;
 
 			case TOKEN_TYPE_CASE:
 			case TOKEN_TYPE_ELSEIF:
 			case TOKEN_TYPE_IF:
-			case TOKEN_TYPE_SWITCH:
 			case TOKEN_TYPE_WHILE:
 				cat_sprintf(buf, "%s%c%s {%s}\n%s{\n", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, indent(token->lvl));
 				break;
 
 			case TOKEN_TYPE_REGEX:
-				cat_sprintf(buf, "%s%c%s {%s} {%s} {%s}", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, token->regex->str, token->regex->bod);
+				cat_sprintf(buf, "%s%c%s {%s} {%s}\n%s{\n%s%s\n%s}", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, token->regex->str, indent(token->lvl), indent(token->lvl + 1), token->regex->bod, indent(token->lvl));
 				break;
 
 			case TOKEN_TYPE_END:
@@ -1081,10 +1085,96 @@ char *write_script(struct session *ses, struct scriptroot *root)
 	return buf;
 }
 
+char *view_script(struct session *ses, struct scriptroot *root)
+{
+	struct scriptnode *token;
+	static char buf[STRING_SIZE];
+
+	token = root->next;
+
+	buf[0] = 0;
+
+	while (token)
+	{
+		switch (token->type)
+		{
+			case TOKEN_TYPE_STRING:
+				cat_sprintf(buf, "%s" COLOR_STRING "%s", indent(token->lvl), token->str);
+				break;
+
+			case TOKEN_TYPE_BREAK:
+			case TOKEN_TYPE_CONTINUE:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name);
+				break;
+
+			case TOKEN_TYPE_RETURN:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s" COLOR_STRING "%s%s", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, *token->str ? " " : "", token->str);
+				break;
+
+			case TOKEN_TYPE_COMMAND:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_COMMAND "%s" COLOR_STRING "%s%s", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, *token->str ? " " : "", token->str);
+				break;
+
+			case TOKEN_TYPE_ELSE:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s\n%s" COLOR_BRACE "{\n", indent(token->lvl), gtd->tintin_char, token->str, indent(token->lvl));
+				break;
+
+			case TOKEN_TYPE_DEFAULT:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s\n%s" COLOR_BRACE "{\n", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, indent(token->lvl));
+				break;
+
+			case TOKEN_TYPE_FOREACH:
+			case TOKEN_TYPE_LOOP:
+			case TOKEN_TYPE_PARSE:
+			case TOKEN_TYPE_SWITCH:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s " COLOR_STRING "%s\n%s" COLOR_BRACE "{\n", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->data->hlt, indent(token->lvl));
+				break;
+
+			case TOKEN_TYPE_CASE:
+			case TOKEN_TYPE_ELSEIF:
+			case TOKEN_TYPE_IF:
+			case TOKEN_TYPE_WHILE:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s " COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "}" COLOR_STRING "\n%s" COLOR_BRACE "{\n", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, indent(token->lvl));
+				break;
+
+			case TOKEN_TYPE_REGEX:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STATEMENT "%s " COLOR_BRACE "{" COLOR_STRING "%s" COLOR_BRACE "} {" COLOR_STRING "%s" COLOR_BRACE "}\n%s{\n%s" COLOR_STRING "%s\n" COLOR_BRACE "%s}", indent(token->lvl), gtd->tintin_char, command_table[token->cmd].name, token->str, token->regex->str, indent(token->lvl), indent(token->lvl + 1), token->regex->bod, indent(token->lvl));
+				break;
+
+			case TOKEN_TYPE_END:
+				cat_sprintf(buf, "\n%s" COLOR_BRACE "}" COLOR_STRING, indent(token->lvl));
+				break;
+
+			case TOKEN_TYPE_SESSION:
+				cat_sprintf(buf, "%s" COLOR_TINTIN "%c" COLOR_STRING "%s", indent(token->lvl), gtd->tintin_char, token->str);
+				break;
+
+			default:
+				tintin_printf2(ses, "#ERROR: UNKNOWN TOKEN TYPE: %d", token->type);
+				break;
+		}
+
+		if (token->next && token->lvl == token->next->lvl)
+		{
+			strcat(buf, COLOR_SEPARATOR ";\n");
+		}
+
+		token = token->next;
+	}
+
+	while (root->next)
+	{
+		deltoken(root, root->next);
+	}
+
+	free(root);
+
+	return buf;
+}
+
 struct session *script_driver(struct session *ses, int list, char *str)
 {
 	struct scriptroot *root;
-	int debug;
 
 	push_call("script_driver(%p,%d,%p)",ses,list,str);
 
@@ -1094,21 +1184,22 @@ struct session *script_driver(struct session *ses, int list, char *str)
 	root->list = list;
 	root->local = init_list(ses, LIST_VARIABLE, LIST_SIZE);
 
-	debug = HAS_BIT(ses->list[list]->flags, LIST_FLAG_DEBUG);
+	gtd->level->input += list != LIST_COMMAND;
 
-	gtd->debug_level += debug;
-	gtd->input_level += list != LIST_COMMAND;
-
-	script_stack[++script_index] = root;
+	gtd->script_stack[++gtd->script_index] = root;
 
 	tokenize_script(root, 0, str);
 
 	ses = (struct session *) parse_script(root, 0, root->next, root->prev);
 
-	script_index--;
+	if (--gtd->script_index == 0)
+	{
+		DEL_BIT(gtd->flags, TINTIN_FLAG_LOCAL);
 
-	gtd->debug_level -= debug;
-	gtd->input_level -= list != LIST_COMMAND;
+//		gtd->memory_index->len = 0;
+	}
+
+	gtd->level->input -= list != LIST_COMMAND;
 
 	while (root->prev)
 	{
@@ -1118,6 +1209,11 @@ struct session *script_driver(struct session *ses, int list, char *str)
 	free_list(root->local);
 	free(root);
 
+	if (HAS_BIT(ses->flags, SES_FLAG_CLOSED))
+	{
+		pop_call();
+		return gtd->ses;
+	}
 	pop_call();
 	return ses;
 }
@@ -1135,3 +1231,16 @@ char *script_writer(struct session *ses, char *str)
 
 	return write_script(ses, root);
 }
+
+char *script_viewer(struct session *ses, char *str)
+{
+	struct scriptroot *root;
+
+	root = (struct scriptroot *) calloc(1, sizeof(struct scriptroot));
+
+	root->ses = ses;
+
+	tokenize_script(root, 1, str);
+
+	return view_script(ses, root);
+}
