@@ -1,12 +1,11 @@
 /******************************************************************************
-*   TinTin++                                                                  *
-*   Copyright (C) 2004 (See CREDITS file)                                     *
+*   This file is part of TinTin++                                             *
 *                                                                             *
-*   This program is protected under the GNU GPL (See COPYING)                 *
+*   Copyright 2004-2020 Igor van den Hoven                                    *
 *                                                                             *
-*   This program is free software; you can redistribute it and/or modify      *
+*   TinTin++ is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by      *
-*   the Free Software Foundation; either version 2 of the License, or         *
+*   the Free Software Foundation; either version 3 of the License, or         *
 *   (at your option) any later version.                                       *
 *                                                                             *
 *   This program is distributed in the hope that it will be useful,           *
@@ -15,103 +14,185 @@
 *   GNU General Public License for more details.                              *
 *                                                                             *
 *   You should have received a copy of the GNU General Public License         *
-*   along with this program; if not, write to the Free Software               *
-*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
-*******************************************************************************/
+*   along with TinTin++.  If not, see https://www.gnu.org/licenses.           *
+******************************************************************************/
 
 /******************************************************************************
-*                (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t                 *
+*                               T I N T I N + +                               *
 *                                                                             *
 *                      coded by Igor van den Hoven 2004                       *
 ******************************************************************************/
 
 #include "tintin.h"
 
-void save_pos(struct session *ses)
+void init_pos(struct session *ses, int row, int col)
 {
-	printf("\e7"); 
+	push_call("init_pos(%p)",ses);
 
-	ses->sav_row = ses->cur_row;
-	ses->sav_col = ses->cur_col;
+	goto_pos(ses, row, col);
+
+	gtd->screen->sav_row[0] = ses->cur_row;
+	gtd->screen->sav_col[0] = ses->cur_col;
+
+	gtd->screen->sav_lev = 1;
+
+	pop_call();
+	return;
 }
 
+void save_pos(struct session *ses)
+{
+	push_call("save_pos(%p)",ses);
+
+	gtd->screen->sav_row[gtd->screen->sav_lev] = ses->cur_row;
+	gtd->screen->sav_col[gtd->screen->sav_lev] = ses->cur_col;
+
+	if (gtd->screen->sav_lev < STACK_SIZE)
+	{
+		gtd->screen->sav_lev++;
+	}
+	else
+	{
+		syserr_printf(ses, "sav_lev++ above 1000.");
+	}
+
+	if (!HAS_BIT(gtd->flags, TINTIN_FLAG_HIDDENCURSOR))
+	{
+		print_stdout("\e[?25l");
+	}
+
+	pop_call();
+	return;
+}
 
 void restore_pos(struct session *ses)
 {
-	printf("\e8\e8"); 
+	if (gtd->screen->sav_lev > 0)
+	{
+		gtd->screen->sav_lev--;
+	}
+	else
+	{
+		gtd->screen->sav_lev = 0;
+		syserr_printf(ses, "sav_lev-- below 0.");
+	}
 
-	ses->cur_row = ses->sav_row;
-	ses->cur_col = ses->sav_col;
+	if (gtd->screen->sav_lev == 1 /* gtd->screen->sav_row[gtd->screen->sav_lev] == inputline_cur_row()*/ /*gtd->screen->rows*/)
+	{
+		goto_pos(ses, inputline_cur_row(), inputline_cur_col());
+	}
+	else
+	{
+		goto_pos(ses, gtd->screen->sav_row[gtd->screen->sav_lev], gtd->screen->sav_col[gtd->screen->sav_lev]);
+	}
+
+	if (!HAS_BIT(gtd->flags, TINTIN_FLAG_HIDDENCURSOR))
+	{
+		print_stdout("\e[?25h");
+	}
 }
 
-void goto_rowcol(struct session *ses, int row, int col)
+void goto_pos(struct session *ses, int row, int col)
 {
-	printf("\e[%d;%dH", row, col);
+	print_stdout("\e[%d;%dH", row, col);
 
+	if (col < 1)
+	{
+		print_stdout("debug: goto_pos(%d,%d)\n",row,col);
+	}
 	ses->cur_row = row;
+	ses->cur_col = col;
 }
 
+void erase_cols(int cnt)
+{
+	if (cnt)
+	{
+		print_stdout("\e[%dX", cnt);
+	}
+}
+
+void erase_row(struct session *ses)
+{
+	if (ses->wrap == gtd->screen->cols || ses->cur_row == gtd->screen->rows)
+	{
+		print_stdout("\e[2K");
+	}
+	else
+	{
+		save_pos(ses);
+
+		goto_pos(ses, ses->cur_row, ses->split->top_col);
+
+		erase_cols(ses->wrap);
+
+		restore_pos(ses);
+	}
+}
+
+void erase_lines(struct session *ses, int rows)
+{
+	print_stdout("\e[%dM", rows);
+}
 
 void erase_toeol(void)
 {
-	printf("\e[K");
+	print_stdout("\e[K");
 }
 
 /*
-	save cursor, goto top row, delete (bot - top) rows, restore cursor
+	unused
 */
 
-void erase_screen(struct session *ses)
+void reset(struct session *ses)
 {
-	printf("\e7\e[%d;1H\e[%dM\e8", ses->top_row, ses->bot_row - ses->top_row);
+	ses->cur_row = 1;
+	ses->cur_col = 1;
 
-	ses->sav_row = ses->cur_row;
-	ses->sav_col = ses->cur_col;
-}
-
-/*
-	doesn't do much
-*/
-
-void reset(void)
-{
-	printf("%cc", ESCAPE);
+	print_stdout("\ec");
 }
 
 
 void scroll_region(struct session *ses, int top, int bot)
 {
-	printf("%c[%d;%dr", ESCAPE, top, bot);
+	push_call("scroll_region(%p,%d,%d)",ses,top,bot);
 
-	ses->top_row = top;
-	ses->bot_row = bot;
+	print_stdout("\e[%d;%dr", top, bot);
+
+	ses->split->top_row = top;
+	ses->split->bot_row = bot;
+
+	check_all_events(ses, SUB_ARG, 0, 4, "VT100 SCROLL REGION", ntos(top), ntos(bot), ntos(gtd->screen->rows), ntos(gtd->screen->cols), ntos(get_scroll_cols(ses)));
+
+	pop_call();
+	return;
 }
 
-
 void reset_scroll_region(struct session *ses)
 {
 	if (ses == gtd->ses)
 	{
-		printf("%c[r", ESCAPE);
+		print_stdout("\e[r");
 	}
-	ses->top_row = 1;
-	ses->bot_row = ses->rows;
+	ses->split->top_row = 1;
+	ses->split->top_col = 1;
+	ses->split->bot_row = gtd->screen->rows;
+	ses->split->bot_col = gtd->screen->cols;
 }
 
 
 int skip_vt102_codes(char *str)
 {
-	int skip;
+	int skip = 0;
 
 	push_call("skip_vt102_codes(%p)",str);
 
 	switch (str[0])
 	{
-		case   5:   /* ENQ */
 		case   7:   /* BEL */
 		case   8:   /* BS  */
-	/*	case   9: *//* HT  */
-	/*	case  10: *//* LF  */
+	//	case   9:      HT
+	//	case  10:      LF
 		case  11:   /* VT  */
 		case  12:   /* FF  */
 		case  13:   /* CR  */
@@ -121,13 +202,28 @@ int skip_vt102_codes(char *str)
 		case  19:   /* DC3 */
 		case  24:   /* CAN */
 		case  26:   /* SUB */
-		case 127:   /* DEL */
 			pop_call();
 			return 1;
 
 		case  27:   /* ESC */
 			break;
 
+		case  28: // HTML_OPEN
+			for (skip = 1 ; str[skip] ; skip++)
+			{
+				if (str[skip] == 30) // HTML_CLOSE
+				{
+					pop_call();
+					return skip + 1;
+				}
+			}
+			pop_call();
+			return 0;
+
+		case 127:   /* DEL */
+			pop_call();
+			return 1;
+
 		default:
 			pop_call();
 			return 0;
@@ -149,6 +245,10 @@ int skip_vt102_codes(char *str)
 		case ']':
 			switch (str[2])
 			{
+				case 0:
+					pop_call();
+					return 2;
+
 				case 'P':
 					for (skip = 3 ; skip < 10 ; skip++)
 					{
@@ -162,7 +262,17 @@ int skip_vt102_codes(char *str)
 
 				case 'R':
 					pop_call();
-					return 3;
+					return str[3] ? 3 : 2;
+
+				default:
+					for (skip = 2 ; str[skip] ; skip++)
+					{
+						if (str[skip] == '\a' || (str[skip] == '\e' && str[skip+1] == '\\'))
+						{
+							pop_call();
+							return skip + 1;
+						}
+					}
 			}
 			pop_call();
 			return 2;
@@ -196,76 +306,44 @@ int skip_vt102_codes(char *str)
 	return skip;
 }
 
-int skip_escaped_vt102_codes(char *str)
+int skip_one_char(struct session *ses, char *str, int *width)
 {
 	int skip;
 
-	if (str[0] != '\\' || str[1] != 'e')
+	*width = 0;
+
+	if (*str)
 	{
-		return 0;
-	}
+		skip = skip_vt102_codes(str);
 
-	switch (str[2])
-	{
-		case '\0':
-			return 2;
-
-		case '%':
-		case '#':
-		case '(':
-		case ')':
-			return str[2] ? 4 : 3;
-
-		case ']':
-			switch (str[3])
-			{
-				case 'P':
-					for (skip = 4 ; skip < 11 ; skip++)
-					{
-						if (str[skip] == 0)
-						{
-							break;
-						}
-					}
-					return skip;
-
-				case 'R':
-					return 4;
-			}
-			return 3;
-
-		case '[':
-			break;
-
-		default:
-			return 3;
-	}
-
-	for (skip = 3 ; str[skip] != 0 ; skip++)
-	{
-		if (isalpha((int) str[skip]))
+		if (skip)
 		{
-			return skip + 1;
+			return skip;
 		}
 
-		switch (str[skip])
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_EUC) && is_euc_head(ses, str))
 		{
-			case '@':
-			case '`':
-			case ']':
-				return skip + 1;
+			return get_euc_width(ses, str, width);
 		}
+
+		if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(str))
+		{
+			return get_utf8_width(str, width);
+		}
+
+		*width = 1;
+		return 1;
 	}
-	return skip;
+	return 0;
 }
 
-int find_non_color_codes(char *str)
+int find_color_code(char *str)
 {
 	int skip;
 
 	switch (str[0])
 	{
-		case  27:   /* ESC */
+		case  ASCII_ESC:
 			break;
 
 		default:
@@ -301,6 +379,113 @@ int find_non_color_codes(char *str)
 	return 0;
 }
 
+int find_escaped_color_code(char *str)
+{
+	int skip;
+
+	switch (str[0])
+	{
+		case  '\\':
+			break;
+
+		default:
+			return 0;
+	}
+
+	switch (str[1])
+	{
+		case 'e':
+			break;
+		default:
+			return 0;
+	}
+
+	switch (str[2])
+	{
+		case '[':
+			break;
+
+		default:
+			return 0;
+	}
+
+	for (skip = 3 ; str[skip] != 0 ; skip++)
+	{
+		switch (str[skip])
+		{
+			case 'm':
+				return skip + 1;
+			case '@':
+			case '`':
+			case ']':
+				return 0;
+		}
+
+		if (isalpha((int) str[skip]))
+		{
+			return 0;
+		}
+	}
+	return 0;
+}
+
+int find_secure_color_code(char *str)
+{
+	int skip, valid = 1;
+
+	switch (str[0])
+	{
+		case  ASCII_ESC:   /* ESC */
+			break;
+
+		default:
+			return 0;
+	}
+
+	switch (str[1])
+	{
+		case '[':
+			break;
+
+		default:
+			return 0;
+	}
+
+	for (skip = 2 ; str[skip] != 0 ; skip++)
+	{
+		switch (str[skip])
+		{
+			case 'm':
+				if (valid)
+				{
+					return skip + 1;
+				}
+				return 0;
+
+			case ';':
+				valid = 0;
+				break;
+
+			case '0':
+			case '1':
+			case '2':
+			case '3':
+			case '4':
+			case '5':
+			case '6':
+			case '7':
+			case '8':
+			case '9':
+				valid = 1;
+				break;
+
+			default:
+				return 0;
+		}
+	}
+	return 0;
+}
+
 
 int skip_vt102_codes_non_graph(char *str)
 {
@@ -308,24 +493,23 @@ int skip_vt102_codes_non_graph(char *str)
 
 	switch (str[skip])
 	{
-		case   5:   /* ENQ */
-		case   7:   /* BEL */
-	/*	case   8: *//* BS  */
-	/*	case   9: *//* HT  */
-	/*	case  10: *//* LF  */
-		case  11:   /* VT  */
-		case  12:   /* FF  */
-		case  13:   /* CR  */
-		case  14:   /* SO  */
-		case  15:   /* SI  */
-		case  17:   /* DC1 */
-		case  19:   /* DC3 */
-		case  24:   /* CAN */
-		case  26:   /* SUB */
-		case 127:   /* DEL */
+		case   7:   // BEL
+//		case   8:   // BS  
+//		case   9:   // HT  
+//		case  10:   // LF  
+		case  11:   // VT  
+		case  12:   // FF  
+		case  13:   // CR  
+		case  14:   // SO  
+		case  15:   // SI  
+		case  17:   // DC1 
+		case  19:   // DC3 
+		case  24:   // CAN 
+		case  26:   // SUB 
+		case 127:   // DEL 
 			return 1;
 
-		case  27:   /* ESC */
+		case  27:   // ESC 
 			break;
 
 		default:
@@ -370,7 +554,7 @@ int skip_vt102_codes_non_graph(char *str)
 				case 'R':
 					return 3;
 			}
-			return 2;
+			return 0;
 
 		case '[':
 			break;
@@ -399,9 +583,7 @@ int skip_vt102_codes_non_graph(char *str)
 	return 0;
 }
 
-
-
-void strip_vt102_codes(char *str, char *buf)
+int strip_vt102_codes(char *str, char *buf)
 {
 	char *pti, *pto;
 
@@ -421,6 +603,8 @@ void strip_vt102_codes(char *str, char *buf)
 		}
 	}
 	*pto = 0;
+
+	return pto - buf;
 }
 
 
@@ -514,10 +698,13 @@ char *strip_vt102_strstr(char *str, char *buf, int *len)
 
 // mix old and str, then copy compressed color string to buf which can point to old.
 
-void get_color_codes(char *old, char *str, char *buf)
+void get_color_codes(char *old, char *str, char *buf, int flags)
 {
 	char *pti, *pto, col[100], tmp[BUFFER_SIZE];
-	int len, vtc, fgc, bgc, cnt, rgb[6];
+	int len, vtc, fgc, bgc, cnt;
+	int rgb[6] = { 0, 0, 0, 0, 0, 0 };
+
+	push_call("get_color_codes(%p,%p,%p,%d)",old,str,buf,flags);
 
 	pto = tmp;
 
@@ -525,7 +712,7 @@ void get_color_codes(char *old, char *str, char *buf)
 
 	while (*pti)
 	{
-		while ((len = find_non_color_codes(pti)) != 0)
+		while ((len = find_color_code(pti)) != 0)
 		{
 			memcpy(pto, pti, len);
 			pti += len;
@@ -542,13 +729,23 @@ void get_color_codes(char *old, char *str, char *buf)
 
 	while (*pti)
 	{
-		while ((len = find_non_color_codes(pti)) != 0)
+		while ((len = find_color_code(pti)) != 0)
 		{
 			memcpy(pto, pti, len);
 			pti += len;
 			pto += len;
 		}
 
+		if (!HAS_BIT(flags, GET_ALL))
+		{
+			break;
+		}
+
+		if (*pti == '\n')
+		{
+			break;
+		}
+
 		if (*pti)
 		{
 			pti++;
@@ -560,6 +757,7 @@ void get_color_codes(char *old, char *str, char *buf)
 	if (strlen(tmp) == 0)
 	{
 		buf[0] = 0;
+		pop_call();
 		return;
 	}
 
@@ -573,7 +771,7 @@ void get_color_codes(char *old, char *str, char *buf)
 	{
 		switch (*pti)
 		{
-			case 27:
+			case ASCII_ESC:
 				pti += 2;
 
 				if (pti[-1] == 'm')
@@ -795,12 +993,15 @@ void get_color_codes(char *old, char *str, char *buf)
 	}
 
 	strcat(buf, "m");
+
+	pop_call();
+	return;
 }
 
 int strip_vt102_strlen(struct session *ses, char *str)
 {
 	char *pti;
-	int i = 0;
+	int w, i = 0;
 
 	pti = str;
 
@@ -813,67 +1014,88 @@ int strip_vt102_strlen(struct session *ses, char *str)
 			continue;
 		}
 
-		if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (*pti & 192) == 192)
+		if (*pti == '\t')
 		{
+			i += ses->tab_width - i % ses->tab_width;
 			pti++;
+		}
+		else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(pti))
+		{
+			pti += get_utf8_width(pti, &w);
 
-			while ((*pti & 192) == 128)
-			{
-				pti++;
-			}
+			i += w;
 		}
 		else
 		{
 			pti++;
+			i++;
 		}
-		i++;
 	}
 	return i;
 }
 
-int strip_color_strlen(struct session *ses, char *str)
+
+int strip_vt102_width(struct session *ses, char *str, int *lines)
 {
 	char *pti;
-	int i = 0;
+	int w, i = 0, max = 0;
+
+	*lines = 1;
 
 	pti = str;
 
 	while (*pti)
 	{
-		if (pti[0] == '<' && isalnum((int) pti[1]) && isalnum((int) pti[2]) && isalnum((int) pti[3]) && pti[4] == '>')
+		if (*pti == '\n')
 		{
-			pti += 5;
-			continue;
+			*lines += 1;
+
+			if (i > max)
+			{
+				max = i;
+				i = 0;
+			}
 		}
 
 		if (skip_vt102_codes(pti))
 		{
 			pti += skip_vt102_codes(pti);
+
 			continue;
 		}
 
-		if (skip_escaped_vt102_codes(pti))
-		{
-			pti += skip_escaped_vt102_codes(pti);
-			continue;
-		}
-
-		if (HAS_BIT(ses->flags, SES_FLAG_UTF8) && (*pti & 192) == 192)
+		if (*pti == '\t')
 		{
+			i += ses->tab_width - i % ses->tab_width;
 			pti++;
+		}
+		else if (HAS_BIT(ses->charset, CHARSET_FLAG_UTF8) && is_utf8_head(pti))
+		{
+			pti += get_utf8_width(pti, &w);
 
-			while ((*pti & 192) == 128)
-			{
-				pti++;
-			}
+			i += w;
 		}
 		else
 		{
 			pti++;
+			i++;
 		}
-		i++;
 	}
-	return i;
+
+	if (i > max)
+	{
+		return i;
+	}
+	return max;
+}
+
+int strip_color_strlen(struct session *ses, char *str)
+{
+	char buf[BUFFER_SIZE];
+
+	substitute(ses, str, buf, SUB_ESC|SUB_COL);
+
+	return strip_vt102_strlen(ses, buf);
 }
 
 int interpret_vt102_codes(struct session *ses, char *str, int real)
@@ -890,8 +1112,17 @@ int interpret_vt102_codes(struct session *ses, char *str, int real)
 		case  27:   /* ESC */
 			break;
 
+		case 11:    /* VT  */
+			ses->cur_row = UMIN(gtd->screen->rows, ses->cur_row + 1);
+			return TRUE;
+
+		case 12:    /* FF  */
+			ses->cur_row = UMIN(gtd->screen->rows, ses->cur_row + 1);
+			return TRUE;
+
 		case  13:   /* CR  */
 			ses->cur_col = 1;
+			return TRUE;
 
 		default:
 			return TRUE;
@@ -917,16 +1148,16 @@ int interpret_vt102_codes(struct session *ses, char *str, int real)
 			return TRUE;
 
 		case 'D':
-			ses->cur_row = URANGE(1, ses->cur_row + 1, ses->rows);
+			ses->cur_row = URANGE(1, ses->cur_row + 1, gtd->screen->rows);
 			return TRUE;
 
 		case 'E':
-			ses->cur_row = URANGE(1, ses->cur_row + 1, ses->rows);
+			ses->cur_row = URANGE(1, ses->cur_row + 1, gtd->screen->rows);
 			ses->cur_col = 1;
 			return TRUE;
 
 		case 'M':
-			ses->cur_row = URANGE(1, ses->cur_row - 1, ses->rows);
+			ses->cur_row = URANGE(1, ses->cur_row - 1, gtd->screen->rows);
 			return TRUE;
 
 		case '[':
@@ -1001,16 +1232,16 @@ int interpret_vt102_codes(struct session *ses, char *str, int real)
 				break;
 
 			case 'r':
-				if (sscanf(data, "%d;%d", &ses->top_row, &ses->bot_row) != 2)
+				if (sscanf(data, "%d;%d", &ses->split->top_row, &ses->split->bot_row) != 2)
 				{
-					if (sscanf(data, "%d", &ses->top_row) != 1)
+					if (sscanf(data, "%d", &ses->split->top_row) != 1)
 					{
-						ses->top_row = 1;
-						ses->bot_row = ses->rows;
+						ses->split->top_row = 1;
+						ses->split->bot_row = gtd->screen->rows;
 					}
 					else
 					{
-						ses->bot_row = ses->rows;
+						ses->split->bot_row = gtd->screen->rows;
 					}
 				}
 				ses->cur_row = 1;
@@ -1038,14 +1269,14 @@ int interpret_vt102_codes(struct session *ses, char *str, int real)
 
 		if (isalpha((int) str[skip]))
 		{
-			ses->cur_row = URANGE(1, ses->cur_row, ses->rows);
+			ses->cur_row = URANGE(1, ses->cur_row, gtd->screen->rows);
 
-			ses->cur_col = URANGE(1, ses->cur_col, ses->cols + 1);
+			ses->cur_col = URANGE(1, ses->cur_col, gtd->screen->cols + 1);
 
 
-			ses->top_row = URANGE(1, ses->top_row, ses->rows);
+			ses->split->top_row = URANGE(1, ses->split->top_row, gtd->screen->rows);
 
-			ses->bot_row = ses->bot_row ? URANGE(1, ses->bot_row, ses->rows) : ses->rows;
+			ses->split->bot_row = ses->split->bot_row ? URANGE(1, ses->split->bot_row, gtd->screen->rows) : gtd->screen->rows;
 			
 			return TRUE;
 		}
